/*
    File:               argonFile.h
    Description:        File I/O handling classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         20.11.2003
    last updated:       11.11.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonFile_h
#define __argonFile_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include <commdlg.h>
    #include "argonBase.h"
    #include "argonString.h"
    #include "argonInstance.h"
    #include "argonErrors.h"
    #include "argonMoreMessageBox.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




#ifndef OFN_DONTADDTORECENT
#define OFN_DONTADDTORECENT          0x02000000
#endif


enum file_t {
    FileT_File              = 0,
    FileT_StioFile,
    FileT_ResourceFile,
    FileT_MemFile,

    _pad_file_t = 0x7fffffff
};

struct FileStatus
{
    FILETIME m_created;         // creation date/time of file
    FILETIME m_modified;        // last modification date/time of file
    FILETIME m_lastAccess;      // last access date/time of file
    UINT64   m_size;            // logical size of file in bytes
    DWORD    m_attribute;       // combination of #File::Attribute# enum values
    String   m_strFullFile;     // absolute path name

    FileStatus() { Empty(); }
    void Empty()
        { FILETIME zero = {0, 0}; m_created = zero; m_modified = zero; m_lastAccess = zero; m_size = 0; m_attribute = 0; m_strFullFile.Empty(); }
};

typedef retval_t<_size_t_>  _retread_t_;
typedef retval_t<INT64>     _retseek_t_;

class File
{
    friend class File;

    public :
        // one of each group of flags must be specified
        enum OpenFlags {
            accessRead =          0x00000000,
            accessWrite =         0x00000001,
            accessReadWrite =     0x00000002,
            shareExclusive =      0x00000000,
            shareDenyWrite =      0x00000010,
            shareDenyRead =       0x00000020,
            shareDenyNone =       0x00000030,
            modeOpen =            0x00000000,
            modeCreate =          0x00000100,
            modeAppend =          0x00000200,
            handleInherit =       0x00000000,
            handleNoInherit =     0x00001000,
            typeText =            0x10000000, // used only by some derived classes, ignored in 'File' class
            typeBinary =          0x20000000  // used only by some derived classes, ignored in 'File' class
        };
        enum SeekPosition {
            begin =               FILE_BEGIN,
            current =             FILE_CURRENT,
            end =                 FILE_END
        };
        enum Attribute {
            normal =              0,
            readonly =            FILE_ATTRIBUTE_READONLY,
            hidden =              FILE_ATTRIBUTE_HIDDEN,
            system =              FILE_ATTRIBUTE_SYSTEM,
            volume =              0x08,
            directory =           FILE_ATTRIBUTE_DIRECTORY,
            archive =             FILE_ATTRIBUTE_ARCHIVE,
            encrypted =           FILE_ATTRIBUTE_ENCRYPTED,
            normal_ =             FILE_ATTRIBUTE_NORMAL,
            temporary =           FILE_ATTRIBUTE_TEMPORARY,
            sparse_file =         FILE_ATTRIBUTE_SPARSE_FILE,
            reparse_point =       FILE_ATTRIBUTE_REPARSE_POINT,
            compressed =          FILE_ATTRIBUTE_COMPRESSED,
            offline =             FILE_ATTRIBUTE_OFFLINE,
            not_content_indexed = FILE_ATTRIBUTE_NOT_CONTENT_INDEXED,
            #ifdef FILE_ATTRIBUTE_INROM         // Windows CE only
            inrom =               FILE_ATTRIBUTE_INROM,
            #endif
            #ifdef FILE_ATTRIBUTE_ROMMODULE     // Windows CE only
            rommodule =           FILE_ATTRIBUTE_ROMMODULE,
            #endif
        };

    protected :
        String m_strFullFile;
        HANDLE m_hFile;
        bool   m_bCloseHFileOnDelete;   // if 'm_hFile' has been set for this object from an already existing file -> file should not be closed by this object
        bool   m_bEof;

        file_t m_FileType;

    public :
        File();
        File( HANDLE hfile );
        File( PCTSTR pstrFile, DWORD openflags );
        virtual ~File();

        file_t getFileType() const { return m_FileType; }

        operator HANDLE() const { return m_hFile; }

        virtual bool Eof() const { return m_bEof; }

        virtual String getFullFile () const;
        virtual String getFullDir  () const { return getFullFile().DirectoryOf(); }
        virtual String getFilename () const { return getFullFile().FilenameOf();  }
        virtual String getFileTitle() const { return getFullFile().FileTitleOf(); }

        virtual reterr_t Open  ( PCTSTR pstrFile, DWORD openflags );
        virtual bool     isOpen() const { return m_hFile != INVALID_HANDLE_VALUE; }
        virtual void     Close ();
        virtual reterr_t Flush ();

        virtual _retread_t_ Read           ( void  *pBuffer, _size_t_ count );
        reterr_t            ReadExactCount ( void  *pBuffer, _size_t_ count );
        reterr_t ReadUChar ( UCHAR  *p ) { return ReadExactCount( (PVOID) p, sizeof(UCHAR ) ); }
        reterr_t ReadChar  ( CHAR   *p ) { return ReadExactCount( (PVOID) p, sizeof(CHAR  ) ); }
        reterr_t ReadWchar ( WCHAR  *p ) { return ReadExactCount( (PVOID) p, sizeof(WCHAR ) ); }
        reterr_t ReadByte  ( BYTE   *p ) { return ReadExactCount( (PVOID) p, sizeof(BYTE  ) ); }
        reterr_t ReadWord  ( WORD   *p ) { return ReadExactCount( (PVOID) p, sizeof(WORD  ) ); }
        reterr_t ReadDWord ( DWORD  *p ) { return ReadExactCount( (PVOID) p, sizeof(DWORD ) ); }
        reterr_t ReadUInt64( UINT64 *p ) { return ReadExactCount( (PVOID) p, sizeof(UINT64) ); }
        _retread_t_         Peek           ( void  *pBuffer, _size_t_ count );
        reterr_t            PeekExactCount ( void  *pBuffer, _size_t_ count );
        reterr_t PeekUChar ( UCHAR  *p ) { return PeekExactCount( (PVOID) p, sizeof(UCHAR ) ); }
        reterr_t PeekChar  ( CHAR   *p ) { return PeekExactCount( (PVOID) p, sizeof(CHAR  ) ); }
        reterr_t PeekWchar ( WCHAR  *p ) { return PeekExactCount( (PVOID) p, sizeof(WCHAR ) ); }
        reterr_t PeekByte  ( BYTE   *p ) { return PeekExactCount( (PVOID) p, sizeof(BYTE  ) ); }
        reterr_t PeekWord  ( WORD   *p ) { return PeekExactCount( (PVOID) p, sizeof(WORD  ) ); }
        reterr_t PeekDWord ( DWORD  *p ) { return PeekExactCount( (PVOID) p, sizeof(DWORD ) ); }
        reterr_t PeekUInt64( UINT64 *p ) { return PeekExactCount( (PVOID) p, sizeof(UINT64) ); }
        virtual reterr_t    Write          ( const void *pBuffer, _size_t_ count );
        virtual PSTR        FgetsA         ( PSTR     pstrBuf, _size_t_ zlenMax );
        virtual PWSTR       FgetsW         ( PWSTR    pstrBuf, _size_t_ zlenMax );
        PTSTR               Fgets          ( PTSTR    pstrBuf, _size_t_ zlenMax );
        reterr_t            FgetstrA       ( StringA& str, bool bRemoveLastNewline = true, bool bAppend = false );
        reterr_t            FgetstrW       ( StringW& str, bool bRemoveLastNewline = true, bool bAppend = false );
        reterr_t            Fgetstr        ( String&  str, bool bRemoveLastNewline = true, bool bAppend = false );
        virtual reterr_t    FputsA         ( PCSTR    pstr );
        virtual reterr_t    FputsW         ( PCWSTR   pstr );
        reterr_t            Fputs          ( PCTSTR   pstr );

            ///<param name="maxTextlines"> Maximum number of text lines to read. <br/>
            ///  0 = no limit </param>
            ///<remarks> You will have to ensure by yourself if you have opened
            ///  the file in text mode or not (see <b>StdioFile</b> class). </remarks>
        virtual reterr_t    readText ( String&  text, _size_t_ maxTextlines = 0, bool bAppend = false );
        virtual reterr_t    readTextA( StringA& text, _size_t_ maxTextlines = 0, bool bAppend = false );
        virtual reterr_t    readTextW( StringW& text, _size_t_ maxTextlines = 0, bool bAppend = false );
            ///<summary> allocates memory and fills it with the data from the
            ///  file beginning at the current file position up to the end-of-file
            ///  or a specified maximum value. </summary>
            ///<param name="bufferSize"> On return will contain the
            ///  number of bytes filled with data. This should also be the
            ///  number of bytes allocated. But if less have been read (e.g. for
            ///  text files), then the size having been read will be returned here. </param>
            ///<param name="maxBytes"> Maximum number of bytes to read. <br/>
            ///  0 = no limit </param>
            ///<returns> Returns a buffer that has been allocated using <b>malloc()</b>
            ///  and therefore can be freed using <b>free()</b>. </returns>
            ///<remarks> This method shall be used mainly with binary type files. <br/>
            ///  IMPORTANT: Because of malloc, we are limited to allocate in
            ///  sizes suitable for malloc. On 32 bit compilers, this is a
            ///  32 bit size value. </remarks>
        virtual PBYTE       readBinary( UINT64& bufferSize, UINT64 maxBytes = 0 );

            ///<summary>
            ///  <b>directaccessReadMemory()</b> returns the pointer at the current
            ///  filepointer position and the remaining size in bytes if <b>this</b>
            ///  File object allows direct memory access of the file's contents.
            ///</summary>
            ///<remarks>
            /// It returns true if <b>this</b> object supports direct memory access
            ///  even if <b>ptr</b> is returned as NULL or <b>count</b> as zero because
            ///  of e.g. EOF reached. <br/>
            /// The filepointer will not be changed in here. <br/>
            ///  But this method *will* update the EOF marker but only if direct memory
            ///  access is possible for <b>this</b> File object. <br/>
            /// If this method returns true and you have read the memory directly,
            ///  just call <b>Seek()</b> to proceed the filepointer after the portion
            ///  that have read.
            ///</remarks>
        virtual bool        directaccessReadMemory( const BYTE *& ptr, _size_t_& count ) { return false; }
            ///<summary>
            ///  <b>directaccessWriteMemory()</b> returns the pointer at the current
            ///  filepointer position for writing the desired number of bytes. <br/>
            ///</summary>
            ///<remarks>
            /// It returns true if <b>this</b> object supports direct memory access
            ///  even if <b>ptr</b> is returned as NULL which indicates that now writing
            ///  is not possible because of e.g. the file cannot be resized as needed. <br/>
            /// The file's internal file-size will be increased if necessary,
            ///  the mem-pointer will be returned, and the filepointer will be moved
            ///  after the desired range of bytes. <br/>
            ///  But if <b>ptr</b> will be returned as NULL, then writing is not possible.
            /// False is returned only if <b>this</b> File object does not support
            ///  direct memory access at all.
            ///</remarks>
        virtual bool        directaccessWriteMemory( BYTE *& ptr, _size_t_ count ) { return false; }

            ///<summary>
            ///  <b>CopyFromFile()</b> reads data from the file <b>f</b>
            ///  at its current filepointer position and writes it into
            ///  <b>this</b> file at our current filepointer position.
            ///</summary>
            ///<param name="f"> The file to read from. </param>
            ///<param name="countToCopy"> number of bytes to copy (exact or at maximum) </param>
            ///<param name="mincountToCopy"> minimum number of bytes to copy.
            ///  If only less bytes can be read, an error will be returned and the
            ///  <b>f</b>'s filepointer position will *not* have changed then. <br/>
            ///  To allow only the exact desired number of bytes to be copied and return
            ///  an error otherwise, set this param to the same value as <i>countToCopy</i>. </param>
            ///<returns> returns the number of bytes copied (successfully read *and* written).
            ///  Use <b>.succeeded()</b> to determine if reading or writing failed. </returns>
            ///<remarks>
            ///  If an error is returned, the read-filepointer position of <b>f</b>
            ///  will not have been changed. <br/>
            ///  The filepointer of <b>this</b> object will be at the position
            ///  as if <b>Write()</b> had been called with the successfully read bytes
            ///  to be copied (get it using <b>_retread_t_::getValue()</b>).
            ///  And the file-size of <b>this</b> object could have been changed
            ///  when a (read) error occured.
            ///</remarks>
        _retread_t_         CopyFromFile( File *f, _size_t_ countToCopy, _size_t_ mincountToCopy = 1 );

        virtual _retseek_t_ Seek( INT64 nOfs, SeekPosition nFrom );
        _retseek_t_ SeekToBegin() { return Seek( 0, begin ); }
        _retseek_t_ SeekToEnd  () { return Seek( 0, end   ); }
        virtual _retseek_t_ Tell() const;
        virtual reterr_t    setLength( UINT64 length );
        virtual INT64       getLength() const;

        virtual reterr_t LockRange  ( UINT64 pos, UINT64 count );
        virtual reterr_t UnlockRange( UINT64 pos, UINT64 count );

        reterr_t getStatus( FileStatus& status ) const;
        static reterr_t getStatus( PCTSTR pstrFile, FileStatus& status );

        static reterr_t createDirectoryPath ( PCTSTR pstrDir, LPSECURITY_ATTRIBUTES lpSecurityAttributes = NULL );
        static reterr_t deleteWholeDirectory( PCTSTR pstrDir );
        static bool     directoryExists( PCTSTR pstrDir, bool bAllowDrivesOnly = true );
        static bool     fileExists( PCTSTR pstrFile );

        static String   browseForDirectory( HWND hwndParent, PCTSTR pstrTitle = NULL, UINT flags = BIF_RETURNONLYFSDIRS );
        static reterr_t browseForDirectory( LPMALLOC pMalloc, HWND hwndParent, PTSTR pstrDir, _size_t_ zlenDir, PCTSTR pstrTitle = NULL, UINT flags = BIF_RETURNONLYFSDIRS, bool *pbInvalidDir = NULL );

        class RestoreSeek
        {
            private :
                File  *m_pFile;
                INT64  m_nSeekPos;

            public :
                RestoreSeek( File& f, bool bStorePosNow = true )
                    { m_pFile = &f; m_nSeekPos = -1; if( bStorePosNow ) StorePos(); }
                ~RestoreSeek() { RestorePos(); }

                bool hasPos() const { return m_nSeekPos >= 0; }

                reterr_t StorePos();
                reterr_t RestorePos();
                void     Dismiss() { m_nSeekPos = -1; }
        };

    protected :
        File( const File& src );
};



class FileFinder
{
    protected :
        HANDLE          m_hFF;
        WIN32_FIND_DATA m_wfd;
        String          m_strFullPath;  // full path where the search is done

    public :
        FileFinder() { m_hFF = INVALID_HANDLE_VALUE; memset( &m_wfd, 0, sizeof(m_wfd) ); }
        FileFinder( PCTSTR pstrWild )
            { m_hFF = INVALID_HANDLE_VALUE; memset( &m_wfd, 0, sizeof(m_wfd) ); FindFirstfile( pstrWild ); }
        virtual ~FileFinder() { Close(); }

    public :
        reterr_t FindFirstfile( PCTSTR pstrWild = NULL );
        reterr_t FindNextfile();
        void     Close();
        bool     isOpen() const { return m_hFF != INVALID_HANDLE_VALUE; }

    public :
        bool   isDots          () const;
        String getFoundFile    () const { return _isstrempty_(m_wfd.cFileName) ? String() : m_strFullPath | m_wfd.cFileName; }
        String getFoundFilename() const { return m_wfd.cFileName; }
        String getSearchPath   () const { return m_strFullPath; }
        DWORD  getAttrib       () const { return m_wfd.dwFileAttributes; }
        bool   isArchive       () const { return ( getAttrib() & FILE_ATTRIBUTE_ARCHIVE    ) != 0; }
        bool   isCompressed    () const { return ( getAttrib() & FILE_ATTRIBUTE_COMPRESSED ) != 0; }
        bool   isDirectory     () const { return ( getAttrib() & FILE_ATTRIBUTE_DIRECTORY  ) != 0; }
        bool   isHidden        () const { return ( getAttrib() & FILE_ATTRIBUTE_HIDDEN     ) != 0; }
        bool   isReadOnly      () const { return ( getAttrib() & FILE_ATTRIBUTE_READONLY   ) != 0; }
        bool   isSystem        () const { return ( getAttrib() & FILE_ATTRIBUTE_SYSTEM     ) != 0; }
        bool   isTemp          () const { return ( getAttrib() & FILE_ATTRIBUTE_TEMPORARY  ) != 0; }
        UINT64 getSizeLarge    () const { ULARGE_INTEGER n; n.LowPart = m_wfd.nFileSizeLow; n.HighPart = m_wfd.nFileSizeHigh; return n.QuadPart; }

    protected :
        FileFinder( const FileFinder& src );
};



class StdioFile : public File
{
    friend class StdioFile;

    protected :
        FILE *m_pStream;
        bool  m_bCloseStreamOnDelete;  // if 'm_pStream' has been set for this object from an already existing file -> file should not be closed by this object

    public :
        StdioFile();
        StdioFile( FILE *stream );
        StdioFile( PCTSTR pstrFile, DWORD openflags );
        virtual ~StdioFile();

        operator FILE*() const { return m_pStream; }

        virtual bool Eof() const { return m_pStream ? (feof(m_pStream) != 0) : false; }

        virtual reterr_t Open ( PCTSTR pstrFile, DWORD openflags );
        virtual void     Close();
        virtual reterr_t Flush();

        virtual _retread_t_ Read  ( void  *pBuffer, _size_t_ count );
        virtual reterr_t    Write ( const void *pBuffer, _size_t_ count );
        virtual PSTR        FgetsA( PSTR   pstrBuf, _size_t_ zlenMax );
        virtual PWSTR       FgetsW( PWSTR  pstrBuf, _size_t_ zlenMax );
        virtual reterr_t    FputsA( PCSTR  pstr );
        virtual reterr_t    FputsW( PCWSTR pstr );

        virtual _retseek_t_ Seek( INT64 nOfs, SeekPosition nFrom );
        virtual _retseek_t_ Tell() const;

        virtual reterr_t LockRange  ( UINT64 pos, UINT64 count ) { return syserr (ERROR_NOT_SUPPORTED); } // not supported in #StdioFile#
        virtual reterr_t UnlockRange( UINT64 pos, UINT64 count ) { return syserr (ERROR_NOT_SUPPORTED); } // not supported in #StdioFile#

    protected :
        StdioFile( const StdioFile& src );
};



class ResourceFile : public File
{
    protected :
        PBYTE    m_pResStart;
        PBYTE    m_pResPtr;
        _size_t_ m_cResData;

    public :
        ResourceFile();
        ResourceFile( UINT resid, PCTSTR pResType );                          // 'pResType' must be name of resource type or a RT_* value (RT_* values used by '::FindResource()')
        ResourceFile( HMODULE hModule, UINT resid, PCTSTR pResType );         // 'pResType' must be name of resource type or a RT_* value (RT_* values used by '::FindResource()')
        ResourceFile( HMODULE hModule, PCTSTR pstrResName, PCTSTR pResType ); // 'pResType' must be name of resource type or a RT_* value (RT_* values used by '::FindResource()')
        virtual ~ResourceFile() { Close(); }

        const BYTE *getResourceMemPtr() const { return m_pResStart; }

        virtual reterr_t Open  ( PCTSTR pstrFile, DWORD openflags ) { return syserr (ERROR_NOT_SUPPORTED); } // Note: Opening from a file is not possible for this resource data access.
        reterr_t Open( UINT resid, PCTSTR pResType );                          // 'pResType' must be name of resource type or a RT_* value (RT_* values used by '::FindResource()')
        reterr_t Open( HMODULE hModule, UINT resid, PCTSTR pResType );         // 'pResType' must be name of resource type or a RT_* value (RT_* values used by '::FindResource()')
        reterr_t Open( HMODULE hModule, PCTSTR pstrResName, PCTSTR pResType ); // 'pResType' must be name of resource type or a RT_* value (RT_* values used by '::FindResource()')
        virtual bool     isOpen() const { return m_pResStart && m_cResData > 0; }
        virtual void     Close ();
        virtual reterr_t Flush ();

        virtual _retread_t_ Read ( void       *pBuffer, _size_t_ count );
        virtual reterr_t    Write( const void *pBuffer, _size_t_ count ) { return syserr (ERROR_NOT_SUPPORTED); } // Resource data cannot be written.

            ///<summary>
            ///  <b>directaccessReadMemory()</b> returns the pointer at the current
            ///  filepointer position and the remaining size in bytes if <b>this</b>
            ///  File object allows direct memory access of the file's contents.
            ///</summary>
            ///<remarks>
            /// It returns true if <b>this</b> object supports direct memory access
            ///  even if <b>ptr</b> is returned as NULL or <b>count</b> as zero because
            ///  of e.g. EOF reached. <br/>
            /// The filepointer will not be changed in here. <br/>
            ///  But this method *will* update the EOF marker but only if direct memory
            ///  access is possible for <b>this</b> File object. <br/>
            /// If this method returns true and you have read the memory directly,
            ///  just call <b>Seek()</b> to proceed the filepointer after the portion
            ///  that have read.
            ///</remarks>
        virtual bool        directaccessReadMemory( const BYTE *& ptr, _size_t_& count );

        virtual _retseek_t_ Seek( INT64 nOfs, SeekPosition nFrom );
        virtual _retseek_t_ Tell() const;
        const BYTE         *TellPtr() const;
        virtual INT64       getLength() const { return m_cResData; } ///<remarks> file pointer will be still on the same location. </remarks> <returns> -1 = error </returns>

    protected :
        ResourceFile( const ResourceFile& src );
};



class MemFile : public File
{
    public :
            ///<remarks> see descriptions to <b>MemFile</b>'s constructor and its
            ///  <i>memflags</i> parameter on description to the different flags. </remarks>
        enum eMemFlags {
            memf_CopyMemory = 0,
            memf_TakeOverMemoryFromMalloc,
            memf_Reference,
        };

    protected :
        PBYTE    m_pMemStart;
        PBYTE    m_pMemPtr;
        _size_t_ m_cMemData;    // desired size of the memory
        _size_t_ m_cAllocated;  // allocated size of the memory (block-size aligned)
        bool     m_bReference;  // if true -> the memory is external. it cannot be reallocated or freed in here and it must exist valid as long as this object is using it
        _size_t_ m_cBlockSize;

    public :
        MemFile();
        MemFile( _size_t_ cBlockSize );
        MemFile( void *pBuffer, _size_t_ cBuffer, eMemFlags memflags = memf_CopyMemory, _size_t_ cBlockSize = 1024 );
        virtual ~MemFile() { Close(); }

        BYTE *getMemoryStartPtr() const { return m_pMemStart; }

        virtual reterr_t Open  ( PCTSTR pstrFile, DWORD openflags ) { return syserr (ERROR_NOT_SUPPORTED); } // Note: Opening from a file is not possible for this class.
        reterr_t CopyMemoryFrom          ( void *pBuffer, _size_t_ cBuffer );
        reterr_t TakeOverMemoryFromMalloc( void *pBuffer, _size_t_ cBuffer );
        reterr_t AttachReference         ( void *pBuffer, _size_t_ cBuffer );
        virtual bool     isOpen() const { return true; }    // the memory file is always open, even if no memory has been allocated yet (even Seek() and Write() will work as usual even in that case)
        virtual void     Close ();
        virtual reterr_t Flush ();

        virtual _retread_t_ Read ( void       *pBuffer, _size_t_ count );
        virtual reterr_t    Write( const void *pBuffer, _size_t_ count );

            ///<summary>
            ///  <b>directaccessReadMemory()</b> returns the pointer at the current
            ///  filepointer position and the remaining size in bytes if <b>this</b>
            ///  File object allows direct memory access of the file's contents.
            ///</summary>
            ///<remarks>
            /// It returns true if <b>this</b> object supports direct memory access
            ///  even if <b>ptr</b> is returned as NULL or <b>count</b> as zero because
            ///  of e.g. EOF reached. <br/>
            /// The filepointer will not be changed in here. <br/>
            ///  But this method *will* update the EOF marker but only if direct memory
            ///  access is possible for <b>this</b> File object. <br/>
            /// If this method returns true and you have read the memory directly,
            ///  just call <b>Seek()</b> to proceed the filepointer after the portion
            ///  that have read.
            ///</remarks>
        virtual bool        directaccessReadMemory( const BYTE *& ptr, _size_t_& count );
            ///<summary>
            ///  <b>directaccessWriteMemory()</b> returns the pointer at the current
            ///  filepointer position for writing the desired number of bytes. <br/>
            ///</summary>
            ///<remarks>
            /// It returns true if <b>this</b> object supports direct memory access
            ///  even if <b>ptr</b> is returned as NULL which indicates that now writing
            ///  is not possible because of e.g. the file cannot be resized as needed. <br/>
            /// The file's internal file-size will be increased if necessary,
            ///  the mem-pointer will be returned, and the filepointer will be moved
            ///  after the desired range of bytes. <br/>
            ///  But if <b>ptr</b> will be returned as NULL, then writing is not possible.
            /// False is returned only if <b>this</b> File object does not support
            ///  direct memory access at all.
            ///</remarks>
        virtual bool        directaccessWriteMemory( BYTE *& ptr, _size_t_ count );

        virtual _retseek_t_ Seek( INT64 nOfs, SeekPosition nFrom );
        virtual _retseek_t_ Tell() const;
        virtual BYTE       *TellPtr() const;
        virtual reterr_t    setLength( UINT64 length );
            ///<remarks> file pointer will be still on the same location. </remarks> <returns> -1 = error </returns>
        virtual INT64       getLength() const { return m_cMemData; }

    protected :
        MemFile( const MemFile& src );
        virtual reterr_t Realloc( _size_t_ newSize );
};



class FilenameDialog
{
    public :
            ///<summary> true = a SaveFileName dialog will be created,
            ///  false = an OpenFileName dialog will be created </summary>
        bool   m_bSaveDialog;
            ///<summary> parent window for the dialog. If NULL, a nice parent window will be looked up. </summary>
        HWND   m_hwndParent;
            ///<summary> the Parameters for the Open/SaveFileName dialog </summary>
        OPENFILENAME m_ofn;
            ///<summary> Title text to display in the caption of the dialog </summary>
        String m_strCaption;
            ///<summary> a string that acts as pseudo-string-list containing all
            ///  the filename filters to use. A pseudo-string-list is a string that
            ///  contains several sub-strings each separated by a | character and
            ///  the list's end is marked using 2 | characters. </summary>
        String m_strFilters;
            ///<summary> initial file to select when dialog opens </summary>
        String m_strInitialFile;
            ///<summary> initial directory to display when dialog opens </summary>
        String m_strInitialDir;
            ///<summary> default extension to use when user did not specify any extension. </summary>
        String m_strDefaultExtension;

            ///<summary> on return, will contain the directory where the resulting files of <b>m_strResultFiles</b> are located in. </summary>
        String m_strResultDirectory;
            ///<summary> on return, will contain a pseudo-string-list with all selected result files (and subdirs), but each *without* path </summary>
        String m_strResultFiles;
            ///<summary> on return, will contain the number of all selected result files </summary>
        int    m_cResultFiles;
            ///<summary> on return, will contain the full path and filename of the first result file returned. </summary>
        String m_strResultFirstFullfilename;

    public :
            ///<summary>
            ///  <b>FilenameDialog()</b> prepares the dialog parameters.
            ///</summary>
            ///<param name="bSaveDialog"> if to create a Save- (true) or OpenFileName (false) dialog. <br/>
            ///  Default = false for an OpenFileName dialog. </param>
            ///<param name="pstrDefExt"> default extension to use when user did not specify any extension. </param>
            ///<param name="pstrInitialFilename"> initial file to select when dialog opens </param>
            ///<param name="dwFlags"> flag's for <b>OPENFILENAME</b> struct's <b>.Flags</b> member, but with one difference: <br/>
            ///  When <b>dwFlags</b> is set to 0, then the <b>m_ofn</b> member's <b>.Flags</b> parameter will be
            ///  initialized with default flags suiting for the desired kind of dialog to be created. </param>
            ///<param name="pstrFilters"> pseudo-string-list containing all the filename filters to use. </param>
            ///<param name="hwndParent"> parent window </param>
        FilenameDialog( bool bSaveDialog = false,
                        PCTSTR pstrDefExt = NULL, PCTSTR pstrInitialFilename = NULL,
                        DWORD dwFlags = 0, PCTSTR pstrFilters = NULL,
                        HWND hwndParent = NULL );

        void PrepareDialogKind( bool bSaveDialog, bool bOverwrite_OFN_Flags = false );

        void AllowMultiselect( bool bMulti    = true  ) { if(  bMulti    ) m_ofn.Flags |= OFN_ALLOWMULTISELECT; else m_ofn.Flags &= ~OFN_ALLOWMULTISELECT; }
        void DontAddToRecent ( bool bAdd      = false ) { if( !bAdd      ) m_ofn.Flags |= OFN_DONTADDTORECENT;  else m_ofn.Flags &= ~OFN_DONTADDTORECENT;  }
        void ShowHelp        ( bool bShowHelp = false ) { if(  bShowHelp ) m_ofn.Flags |= OFN_SHOWHELP;         else m_ofn.Flags &= ~OFN_SHOWHELP;         }

        void AppendAllFilesFilter()
        { m_strFilters.TrimEnd( _T("|") ); m_strFilters += _T("|All Files (*.*)|*.*||"); m_strFilters.TrimStart( _T("|") ); }

        bool DoModal();

        String getResultFilenameAt    ( _idx_ index );
        String getResultFullFilenameAt( _idx_ index );
};



class TextStream
{
    public :
        enum eEncodingType {
            enctAscii                   = 1,
            enctUnicode,
            enctDetectByByteOrderMark   = 0x100,
        };

        enum eByteOrder {
            byorLittleEndian    = 0,
            byorBigEndian,
            byorWindows         = byorLittleEndian
        };

        enum eTextMode {
            txtmdCRLF       = 1,
            txtmdLF,
            txtmdCR,
            txtmdWindows    = txtmdCRLF,
            txtmdUnix       = txtmdLF,
            txtmdMacOS      = txtmdCR,
            txtmdRichedit2  = txtmdCR,
            txtmdDontChange = 0x100,
        };

    protected :
        File          *m_pFile;

    public :
        eEncodingType  m_EncodingType;
        eByteOrder     m_ByteOrder;
        eTextMode      m_TextMode;                  // mode how it should be stored in the file. Used only when writing to the file.
        bool           m_bPrependByteOrderMarker;   // if true, Write() will output a byte order marker first if writing a unicode file.

    public :
        TextStream( File& f,
                    eEncodingType encoding_type = enctDetectByByteOrderMark,
                    eByteOrder byte_order = byorWindows,
                    eTextMode text_mode = txtmdWindows )
            { m_pFile = &f; m_EncodingType = encoding_type; m_ByteOrder = byte_order; m_TextMode = text_mode; m_bPrependByteOrderMarker = true; }

        _retread_t_ ReadA( StringA& str, _size_t_ countChars, eTextMode textmode_for_String = txtmdLF );
        _retread_t_ ReadW( StringW& str, _size_t_ countChars, eTextMode textmode_for_String = txtmdLF );
        _retread_t_ Read ( String&  str, _size_t_ countChars, eTextMode textmode_for_String = txtmdLF );

        _retread_t_ ReadToEndA( StringA& str, eTextMode textmode_for_String = txtmdLF, _size_t_ cIterationChars = 4096 );
        _retread_t_ ReadToEndW( StringW& str, eTextMode textmode_for_String = txtmdLF, _size_t_ cIterationChars = 4096 );
        _retread_t_ ReadToEnd ( String&  str, eTextMode textmode_for_String = txtmdLF, _size_t_ cIterationChars = 4096 );

        reterr_t    WriteA( PCSTR  pstr, _size_t_ countChars );
        reterr_t    WriteW( PCWSTR pstr, _size_t_ countChars );
        reterr_t    Write ( PCTSTR pstr, _size_t_ countChars );
};




}

#endif
