/*
    File:               argonHexView.cpp
    Description:        Generates hex view text from a binary data.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         21.10.2005
    last updated:       03.09.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonHexView_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonHexView.h"

namespace ArgonLibrary
{




HexView::HexView( int numItemsPerLine )
{
    m_bIncludeLineOffsetNrs = true;
    m_bIncludeItems         = true;
    m_bIncludeTextChars     = true;

    m_pString = NULL;
    m_cchString = 0;

    Settings( numItemsPerLine,
              1,
              _T("%08xh: "),
              _T("%02X "),
              _T("   "),
              _T("; "),
              _T('.') );

    m_nextLineOffsetNr      = 0;
    m_pCurrBufferPtr        = NULL;
    m_cRemainingBufferSize  = 0;
}



///<summary>
///  <b>Settings()</b> sets new settings for the hex view.
///  Warning: If you change <b>numBytesPerItem</b>, then you
///  will also have to set an appropriate Item Format string
///  in <b>pstrItemFmt</b>.
///</summary>
///<param name="numItemsPerLine"> number of items per text line to output </param>
///<param name="numBytesPerItem"> number of bytes for one item (normally 1).
///  Maximum = 4 bytes. </param>
///<param name="pstrLineNrFmt"> The format string to use for writing line numbers,
///  if empty string -> no line numbers included. </param>
///<param name="pstrItemFmt"> the format string to use to print one item </param>
///<param name="pstrItemSpacing"> and the string to print when to indent by one
///  item without writing an item </param>
///<param name="pstrSeparator"> the string to append after the items but before
///  the real text characters. if empty string -> no separator </param>
///<param name="chPlaceholder"> this char will be used to display some char in
///  the text char area for characters which normally do not have a
///  displaying character. </param>
///<returns> True on success, false on error. </returns>
bool HexView::Settings( int numItemsPerLine,
                        int numBytesPerItem,
                        PCTSTR pstrLineNrFmt,
                        PCTSTR pstrItemFmt,
                        PCTSTR pstrItemSpacing,
                        PCTSTR pstrSeparator,
                        TCHAR chPlaceholder )
{
    if( numItemsPerLine < 0 ) return false;
    if( numBytesPerItem < 0 ) return false;

    if( pstrLineNrFmt ) m_strLineNrFmt = pstrLineNrFmt;
    if( pstrItemFmt ) m_strItemFmt = pstrItemFmt;
    if( pstrItemSpacing ) m_strItemSpacing = pstrItemSpacing;
    if( pstrSeparator ) m_strSeparator = pstrSeparator;
    if( chPlaceholder ) m_chPlaceholder = chPlaceholder;

    m_numItemsPerLine = numItemsPerLine;
    m_numBytesPerItem = numBytesPerItem;

    _size_t_ cChars = numItemsPerLine * 12 + 128;  // *12 is a big number to be on the safe side. +128 is for line numbers and separators including some safety.

    PTSTR p = (PTSTR) realloc( m_pString, cChars * sizeof(TCHAR) );
    if( !p ) return false;

    m_pString = p;
    m_cchString = cChars;

    return true;
}



///<summary>
///  <b>GetLine()</b> gets an independent line from an
///  arbitrary buffer.
///</summary>
///<param name="pBuffer"> start of the buffer area to parse </param>
///<param name="cBuffer"> number of bytes in the buffer. It is used mainly
///  to determine if the bytes to retrieve from the buffer fit in one
///  line of text or if to retrieve less bytes. </param>
///<param name="pstrAppendNewline"> A string to append at the end.
///  Mainly used to append newline characters. </param>
///<returns> Pointer to the resulting string buffer.
///  NULL if no more bytes to retrieve. </returns>
PCTSTR HexView::GetLine( _idx_ lineOffsetNr, const BYTE *pBuffer, _size_t_ cBuffer, PCTSTR pstrAppendNewline ) const
{
    if( !m_pString || m_cchString <= 0 || !pBuffer || cBuffer <= (_size_t_) 0 ) return NULL;
    if( m_numItemsPerLine <= (_idx_) 0 || m_numBytesPerItem <= 0 || m_strItemFmt.IsEmpty() ) return NULL;

    const BYTE *pbuf = pBuffer;

    PTSTR p = m_pString;
    *p = (TCHAR) 0;
    _size_t_ cch = m_cchString;

    // First the line numbers.

    if( m_bIncludeLineOffsetNrs && !m_strLineNrFmt.IsEmpty() )
    {
        _sprintf_( p, cch, m_strLineNrFmt, lineOffsetNr );
        _size_t_ l = _strlen_( p );
        p += l;
        cch -= l;
    }

    // Now the items.

    _size_t_ bytesPerLine = getBytesPerLine();
    if( cBuffer > bytesPerLine ) cBuffer = bytesPerLine;

    _size_t_ numFullItemsInLine = cBuffer / (_size_t_) m_numBytesPerItem;
    _size_t_ numBytesOfHalfItem = cBuffer - numFullItemsInLine * m_numBytesPerItem;
    _size_t_ numEmptyItems      = m_numItemsPerLine - numFullItemsInLine;
    if( numBytesOfHalfItem > (_size_t_) 0 ) --numEmptyItems;

    if( m_bIncludeItems )
    {
        _size_t_ i;
        for( i = 0 ; i < numFullItemsInLine ; ++i )
        {
            _sprintf_( p, cch, m_strItemFmt, (UINT) GetFullItem(pbuf) );
            _size_t_ l = _strlen_( p );
            p += l;
            cch -= l;
            pbuf += m_numBytesPerItem;
        }

        if( numBytesOfHalfItem > (_size_t_) 0 )
        {
            _sprintf_( p, cch, m_strItemFmt, (UINT) GetItem(pbuf, numBytesOfHalfItem) );
            _size_t_ l = _strlen_( p );
            p += l;
            cch -= l;
        }

        _size_t_ spacingLen = _strlen_( m_strItemSpacing );
        for( i = 0 ; i < numEmptyItems ; ++i )
        {
            _strcpy_( p, m_strItemSpacing, cch );
            p += spacingLen;
            cch -= spacingLen;
        }
    }

    // Separator.

    if( m_bIncludeTextChars && m_bIncludeItems && !m_strSeparator.IsEmpty() )
    {
        _size_t_ sepLen = _strlen_( m_strSeparator );
        _strcpy_( p, m_strSeparator, sepLen + 1 );
        p += sepLen;
    }

    // Finally the text chars.

    if( m_bIncludeTextChars )
    {
        pbuf = pBuffer;

        for( _size_t_ i = 0 ; i < numFullItemsInLine ; ++i )
        {
            TCHAR ch = (TCHAR) GetFullItem(pbuf);
            *(p++) = (WORD) ch >= (WORD) _T(' ') ? ch : m_chPlaceholder;
            pbuf += m_numBytesPerItem;
        }

        if( numBytesOfHalfItem > (_size_t_) 0 )
        {
            TCHAR ch = (TCHAR) GetItem(pbuf, numBytesOfHalfItem);
            *(p++) = (WORD) ch >= (WORD) _T(' ') ? ch : m_chPlaceholder;
        }

        *p = (TCHAR) 0;
    }

    // append

    if( !_isstrempty_(pstrAppendNewline) )
        _strcpy_( p, pstrAppendNewline, _strlen_(pstrAppendNewline) + 1 );

    return m_pString;
}



bool HexView::Restart( _idx_ startlineOffsetNr, const BYTE *pBuffer, _size_t_ cBuffer )
{
    m_nextLineOffsetNr      = startlineOffsetNr;
    m_pCurrBufferPtr        = pBuffer;
    m_cRemainingBufferSize  = cBuffer;

    return m_pCurrBufferPtr && m_cRemainingBufferSize > (_size_t_) 0;
}

///<summary>
///  <b>GetNextLine()</b> gets the next line after having
///  called <b>Restart()</b>.
///</summary>
///<param name="pstrAppendNewline"> A string to append at the end.
///  Mainly used to append newline characters. </param>
///<returns> Pointer to the resulting string buffer.
///  NULL if no more bytes to retrieve. </returns>
PCTSTR HexView::GetNextLine( PCTSTR pstrAppendNewline )
{
    PCTSTR ret = GetLine( m_nextLineOffsetNr, m_pCurrBufferPtr, m_cRemainingBufferSize, pstrAppendNewline );

    _size_t_ bytesPerLine = (_size_t_) getBytesPerLine();
    _size_t_ cntThisLine = m_cRemainingBufferSize < bytesPerLine ? m_cRemainingBufferSize : bytesPerLine;

    m_nextLineOffsetNr      += cntThisLine;
    m_pCurrBufferPtr        += cntThisLine;
    m_cRemainingBufferSize  -= cntThisLine;

    return ret;
}



String HexView::GetFromBuffer( _idx_ startlineOffsetNr, const BYTE *pBuffer, _size_t_ cBuffer, PCTSTR pstrAppendNewline )
{
    String str;
    _size_t_ bytesPerLine = (_size_t_) getBytesPerLine();
    if( bytesPerLine <= (_size_t_) 0 ) return str;

    _idx_ nextLineNr = startlineOffsetNr;
    const BYTE *pbuf = pBuffer;
    _size_t_    cbuf = cBuffer;

    PCTSTR ret;
    while( ret = GetLine( nextLineNr, pbuf, cbuf, pstrAppendNewline ) )
    {
        str += ret;

        _size_t_ cntThisLine = cbuf < bytesPerLine ? cbuf : bytesPerLine;

        nextLineNr += bytesPerLine;
        pbuf += cntThisLine;
        cbuf -= cntThisLine;
    }

    return str;
}



DWORD HexView::GetFullItem( const BYTE *ptr ) const
{
    switch( m_numBytesPerItem )
    {
        case 0  : return 0;
        case 1  : return (DWORD) *ptr;
        case 2  : return (DWORD) *(WORD*)ptr;
        case 3  : return *(DWORD*)ptr & 0xffffff;    // expects CPU to use little-endian
        case 4  : return *(DWORD*)ptr;
        default : return *(DWORD*)ptr;
    }
}

DWORD HexView::GetItem( const BYTE *ptr, int cBuffer ) const
{
    if( cBuffer >= 4 ) return GetFullItem( ptr );
    if( cBuffer <= 0 ) return 0;

    if( cBuffer == 1 && m_numBytesPerItem >= 1 ) return (DWORD) *ptr;
    if( cBuffer == 2 && m_numBytesPerItem >= 2 ) return (DWORD) *(WORD*)ptr;
    if( cBuffer == 3 && m_numBytesPerItem >= 3 )
    {
        DWORD dw = (DWORD) *(WORD*)ptr;
        dw |= (((DWORD) ptr[2]) << 16);
        return dw;
    }

    return GetFullItem( ptr );
}




}
