/*
    File:               argonKernel32Dll.h
    Description:        Parts of the Kernel Dll API.
                        Dynamically loading of some functions of the Kernel Dll API.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         01.12.2005
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonKernel32Dll_h
#define __argonKernel32Dll_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

namespace ArgonLibrary
{




/*
    -----------------------------------------------------------------
    Usage:

    Simply call the functions defined in here their usual way like
    #GetComputerNameEx()# or #GetSystemWow64Directory()#, etc.,
    in your source code as you need it.

    All these functions defined in here are emulating the system behavior
    that way, that they load the functions during runtime from the system
    and only if this did succeed, they will be called.

    The first time you call any of these functions, the library will try
    to be loaded and the function pointers will try to be retrieved.

    The unloading of the library will be done automatically
    during the deleting of the global variables.
    If you included this file in a DLL project, you will not want this
    behavior to happen because it should be prevented to call FreeLibrary()
    when a DLL is detached from a process (where global variables will be
    deleted).
    So in the case of DLL project, it is recommended to export an own
    DLL-deinit-function which you should redirect to a call to
    #Kernel32Dll::UnloadAPI()#. If the user then calls this deinit-function,
    the KERNEL32.DLL will be unloaded and later on, when your DLL is detachted
    from a process, it is harmless to call #Kernel32Dll::UnloadAPI()# again
    because it will not do anything any more.
*/


#define KERNEL32DLL_NAME            "kernel32.dll"


#if defined (__cplusplus)
extern "C" {
#endif

#if !defined( _WIN32_WINNT ) || ( _WIN32_WINNT < 0x0500 )

typedef enum _COMPUTER_NAME_FORMAT {
    ComputerNameNetBIOS,
    ComputerNameDnsHostname,
    ComputerNameDnsDomain,
    ComputerNameDnsFullyQualified,
    ComputerNamePhysicalNetBIOS,
    ComputerNamePhysicalDnsHostname,
    ComputerNamePhysicalDnsDomain,
    ComputerNamePhysicalDnsFullyQualified,
    ComputerNameMax
} COMPUTER_NAME_FORMAT ;

#endif


typedef BOOL (WINAPI *TfnGetComputerNameExA)(
    COMPUTER_NAME_FORMAT NameType,
    LPSTR lpBuffer,
    LPDWORD nSize
);

BOOL WINAPI argonGetComputerNameExA(
    COMPUTER_NAME_FORMAT NameType,
    LPSTR lpBuffer,
    LPDWORD nSize
);

typedef BOOL (WINAPI *TfnGetComputerNameExW)(
    COMPUTER_NAME_FORMAT NameType,
    LPWSTR lpBuffer,
    LPDWORD nSize
);

BOOL WINAPI argonGetComputerNameExW(
    COMPUTER_NAME_FORMAT NameType,
    LPWSTR lpBuffer,
    LPDWORD nSize
);

typedef VOID (WINAPI *TfnGetNativeSystemInfo)(
    LPSYSTEM_INFO lpSystemInfo
);

VOID WINAPI argonGetNativeSystemInfo(
    LPSYSTEM_INFO lpSystemInfo
);

typedef UINT (WINAPI *TfnGetSystemWindowsDirectoryA)(
    LPSTR lpBuffer,
    UINT uSize
);

UINT WINAPI argonGetSystemWindowsDirectoryA(
    LPSTR lpBuffer,
    UINT uSize
);

typedef UINT (WINAPI *TfnGetSystemWindowsDirectoryW)(
    LPWSTR lpBuffer,
    UINT uSize
);

UINT WINAPI argonGetSystemWindowsDirectoryW(
    LPWSTR lpBuffer,
    UINT uSize
);

typedef UINT (WINAPI *TfnGetSystemWow64DirectoryA)(
    LPSTR lpBuffer,
    UINT uSize
);

UINT WINAPI argonGetSystemWow64DirectoryA(
    LPSTR lpBuffer,
    UINT uSize
);

typedef UINT (WINAPI *TfnGetSystemWow64DirectoryW)(
    LPWSTR lpBuffer,
    UINT uSize
);

UINT WINAPI argonGetSystemWow64DirectoryW(
    LPWSTR lpBuffer,
    UINT uSize
);

typedef BOOL (WINAPI *TfnIsProcessorFeaturePresent)(
    DWORD ProcessorFeature
);

BOOL WINAPI argonIsProcessorFeaturePresent(
    DWORD ProcessorFeature
);

typedef BOOL (WINAPI *TfnIsWow64Process)(
    HANDLE hProcess,
    PBOOL Wow64Process
);

BOOL WINAPI argonIsWow64Process(
    HANDLE hProcess,
    PBOOL Wow64Process
);

typedef BOOL (WINAPI *TfnSetComputerNameExA)(
    COMPUTER_NAME_FORMAT NameType,
    LPCSTR lpBuffer
);

BOOL WINAPI argonSetComputerNameExA(
    COMPUTER_NAME_FORMAT NameType,
    LPCSTR lpBuffer
);

typedef BOOL (WINAPI *TfnSetComputerNameExW)(
    COMPUTER_NAME_FORMAT NameType,
    LPCWSTR lpBuffer
);

BOOL WINAPI argonSetComputerNameExW(
    COMPUTER_NAME_FORMAT NameType,
    LPCWSTR lpBuffer
);

#define GetComputerNameExA         ArgonLibrary::argonGetComputerNameExA
#define GetComputerNameExW         ArgonLibrary::argonGetComputerNameExW
#define GetNativeSystemInfo        ArgonLibrary::argonGetNativeSystemInfo
#define GetSystemWindowsDirectoryA ArgonLibrary::argonGetSystemWindowsDirectoryA
#define GetSystemWindowsDirectoryW ArgonLibrary::argonGetSystemWindowsDirectoryW
#define GetSystemWow64DirectoryA   ArgonLibrary::argonGetSystemWow64DirectoryA
#define GetSystemWow64DirectoryW   ArgonLibrary::argonGetSystemWow64DirectoryW
#define IsProcessorFeaturePresent  ArgonLibrary::argonIsProcessorFeaturePresent
#define IsWow64Process             ArgonLibrary::argonIsWow64Process
#define SetComputerNameExA         ArgonLibrary::argonSetComputerNameExA
#define SetComputerNameExW         ArgonLibrary::argonSetComputerNameExW

#ifdef UNICODE
    #define argonGetComputerNameEx          argonGetComputerNameExW
    #define argonGetSystemWindowsDirectory  argonGetSystemWindowsDirectoryW
    #define argonGetSystemWow64Directory    argonGetSystemWow64DirectoryW
    #define argonSetComputerNameEx          argonSetComputerNameExW
#else
    #define argonGetComputerNameEx          argonGetComputerNameExA
    #define argonGetSystemWindowsDirectory  argonGetSystemWindowsDirectoryA
    #define argonGetSystemWow64Directory    argonGetSystemWow64DirectoryA
    #define argonSetComputerNameEx          argonSetComputerNameExA
#endif

#if defined (__cplusplus)
}
#endif


class Kernel32Dll
{
    protected :
        class Unloader
        {
            public :
                Unloader() {  }
                ~Unloader() { if( Kernel32Dll::s_bAutoUnload ) Kernel32Dll::UnloadAPI(); }
        };

        static Unloader s_Unloader;

    public :
        static HMODULE s_hModule;
        static TfnGetComputerNameExA         s_pfnGetComputerNameExA;
        static TfnGetComputerNameExW         s_pfnGetComputerNameExW;
        static TfnGetNativeSystemInfo        s_pfnGetNativeSystemInfo;
        static TfnGetSystemWindowsDirectoryA s_pfnGetSystemWindowsDirectoryA;
        static TfnGetSystemWindowsDirectoryW s_pfnGetSystemWindowsDirectoryW;
        static TfnGetSystemWow64DirectoryA   s_pfnGetSystemWow64DirectoryA;
        static TfnGetSystemWow64DirectoryW   s_pfnGetSystemWow64DirectoryW;
        static TfnIsProcessorFeaturePresent  s_pfnIsProcessorFeaturePresent;
        static TfnIsWow64Process             s_pfnIsWow64Process;
        static TfnSetComputerNameExA         s_pfnSetComputerNameExA;
        static TfnSetComputerNameExW         s_pfnSetComputerNameExW;
        static bool s_bAutoUnload;  // true = automatically unloads the Dll when global objects are destructed.

        Kernel32Dll();

        static void LoadAPI();
        static void UnloadAPI();

    protected :
        static void Nullify();
};




}

#endif
