/*
    File:               argonMinistring.h
    Description:        A small and simple string class.
                        It does not do complex handling and needs only a
                        little set of C runtime and wide character functions.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         27.08.2005
    last updated:       06.07.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonMinistring_h
#define __argonMinistring_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




/*
    miniString usage:
        miniString<char>        ANSI version
        miniString<wchar_t>     wide character (Unicode) version
        miniString<TCHAR>       current character selection

    supported features:
        . Assigning (= operator) from string (ansi or unicode),
            from character (ansi or unicode) or from another miniString.
        . Concatenating (+= operator) with a string (ansi or unicode),
            with a character (ansi or unicode) or with another miniString.
        . Direct access to the buffer and afterwards releasing it.
        . uses heap memory (malloc, realloc, free).
        . needs only C runtime string length, memory allocation, memory copy functions
            as well as WideCharToMultiByte and MultiByteToWideChar.
*/

typedef INT_PTR _idx_;
typedef size_t  _size_t_;

namespace privateMinistring
{
    template< class Tchar > class char_T          { public : typedef wchar_t otherType; enum { isWide = false }; };
    template<>              class char_T<wchar_t> { public : typedef char    otherType; enum { isWide = true  }; };
}

template< class Tchar >
class miniString
{
    protected :
        Tchar *m_pstrString;

    public :
        typedef typename privateMinistring::char_T<Tchar>::otherType TotherChar;
        enum { isWide = privateMinistring::char_T<Tchar>::isWide };

        miniString() { m_pstrString = NULL; }
        miniString( const Tchar      *string ) { m_pstrString = NULL; Copy( string ); }
        miniString( const TotherChar *string ) { m_pstrString = NULL; Copy( string ); }
        miniString( const Tchar       ch     ) { m_pstrString = NULL; Tchar      s[2] = { ch, (Tchar)       0 }; Copy( s ); }
        miniString( const TotherChar  ch     ) { m_pstrString = NULL; TotherChar s[2] = { ch, (TotherChar ) 0 }; Copy( s ); }
        miniString( const miniString& src    ) { m_pstrString = NULL; Copy( src );    }
        ~miniString() { Empty(); }

        operator const Tchar *() const { static Tchar z = (Tchar) 0; return m_pstrString ? m_pstrString : &z; }

        miniString& operator = ( const Tchar      *string ) { Copy( string ); return *this; }
        miniString& operator = ( const TotherChar *string ) { Copy( string ); return *this; }
        miniString& operator = ( const Tchar       ch     ) { Tchar      s[2] = { ch, (Tchar)       0 }; Copy( s ); return *this; }
        miniString& operator = ( const TotherChar  ch     ) { TotherChar s[2] = { ch, (TotherChar ) 0 }; Copy( s ); return *this; }
        miniString& operator = ( const miniString& src    ) { Copy( src );    return *this; }

        miniString& operator +=( const Tchar      *string ) { Cat( string );  return *this; }
        miniString& operator +=( const TotherChar *string ) { Cat( string );  return *this; }
        miniString& operator +=( const Tchar       ch     ) { Tchar      s[2] = { ch, (Tchar)       0 }; Cat( s ); return *this; }
        miniString& operator +=( const TotherChar  ch     ) { TotherChar s[2] = { ch, (TotherChar ) 0 }; Cat( s ); return *this; }
        miniString& operator +=( const miniString& src    ) { Cat( src );     return *this; }

        friend miniString operator +( const miniString& str1, const miniString& str2 )  { miniString s = str1; s += str2; return s; }
        friend miniString operator +( const miniString& str1, const Tchar *psz2 )       { miniString s = str1; s += psz2; return s; }
        friend miniString operator +( const Tchar *psz1, const miniString& str2 )       { miniString s = psz1; s += str2; return s; }
        friend miniString operator +( const miniString& str1, wchar_t ch2 )             { miniString s = str1; s += ch2;  return s; }
        friend miniString operator +( const miniString& str1, char ch2 )                { miniString s = str1; s += ch2;  return s; }
        friend miniString operator +( wchar_t ch1, const miniString& str2 )             { miniString s = ch1;  s += str2; return s; }
        friend miniString operator +( char ch1, const miniString& str2 )                { miniString s = ch1;  s += str2; return s; }

        bool IsEmpty() const { return m_pstrString ? *m_pstrString == 0 : true; }

        void Empty() { if( m_pstrString ) { free( m_pstrString ); m_pstrString = NULL; } }
        bool Copy( const Tchar      *string );
        bool Copy( const TotherChar *string );
        bool Cat ( const Tchar      *string );
        bool Cat ( const TotherChar *string );

        Tchar *GetBuffer() { return m_pstrString; }                         // Note: Will be used by #StringT<>#, too, if miniString is used as baseString class for StringT<>.
        Tchar *GetBuffer( _size_t_ minLength, bool bClearBuffer = false );  // Note: Will be used by #StringT<>#, too, if miniString is used as baseString class for StringT<>.
        bool   ReleaseBuffer();                                             // Note: Will be used by #StringT<>#, too, if miniString is used as baseString class for StringT<>.

        static _size_t_ strlen( PCSTR  string ) { return string ? ::strlen( string ) : (_size_t_) 0; }
        static _size_t_ strlen( PCWSTR string ) { return string ? ::wcslen( string ) : (_size_t_) 0; }

        static bool Convert( PWSTR pstrDest, PCSTR  pstrSrce, int zlenSrce );
        static bool Convert( PSTR  pstrDest, PCWSTR pstrSrce, int zlenSrce );
        static bool Convert( PSTR  pstrDest, PCSTR  pstrSrce, int zlenSrce );
        static bool Convert( PWSTR pstrDest, PCWSTR pstrSrce, int zlenSrce );
};



template< class Tchar >
bool miniString<Tchar>::Copy( const Tchar *string )
{
    _size_t_ len = strlen( string );
    if( len <= 0 ) { Empty(); return true; }

    Tchar *p = (Tchar*) realloc( m_pstrString, (len + 1) * sizeof(Tchar) );
    if( !p ) return false;

    m_pstrString = p;

    return Convert( m_pstrString, string, (int) len + 1 );
}

template< class Tchar >
bool miniString<Tchar>::Copy( const TotherChar *string )
{
    _size_t_ len = strlen( string );
    if( len <= 0 ) { Empty(); return true; }

    Tchar *p = (Tchar*) realloc( m_pstrString, (len + 1) * sizeof(Tchar) );
    if( !p ) return false;

    m_pstrString = p;

    return Convert( m_pstrString, string, (int) len + 1 );
}

template< class Tchar >
bool miniString<Tchar>::Cat( const Tchar *string )
{
    _size_t_ len = strlen( string );
    if( len <= 0 ) return true;

    _size_t_ currlen = strlen( m_pstrString );

    Tchar *p = (Tchar*) realloc( m_pstrString, (currlen + len + 1) * sizeof(Tchar) );
    if( !p ) return false;

    m_pstrString = p;

    return Convert( m_pstrString + currlen, string, (int) len + 1 );
}

template< class Tchar >
bool miniString<Tchar>::Cat( const TotherChar *string )
{
    _size_t_ len = strlen( string );
    if( len <= 0 ) return true;

    _size_t_ currlen = strlen( m_pstrString );

    Tchar *p = (Tchar*) realloc( m_pstrString, (currlen + len + 1) * sizeof(Tchar) );
    if( !p ) return false;

    m_pstrString = p;

    return Convert( m_pstrString + currlen, string, (int) len + 1 );
}

///<param name="minLength"> Minimum length for the buffer (*excluding* the null terminator) </param>
template< class Tchar >
Tchar *miniString<Tchar>::GetBuffer( _size_t_ minLength, bool bClearBuffer )
{
    if( minLength <= 0 ) return NULL;

    _size_t_ currLength = strlen( m_pstrString );
    if( minLength > currLength )
    {
        Tchar *p = (Tchar*) realloc( m_pstrString, (minLength + 1) * sizeof(Tchar) );
        if( !p ) return NULL;

        if( !m_pstrString ) p[0] = (Tchar) 0;

        m_pstrString = p;
        currLength = minLength;
    }

    if( bClearBuffer ) memset( m_pstrString, 0, (currLength + 1) * sizeof(Tchar) );
    else               m_pstrString[currLength] = (Tchar) 0;

    return m_pstrString;
}

template< class Tchar >
bool miniString<Tchar>::ReleaseBuffer()
{
    _size_t_ len = strlen( m_pstrString );
    if( len <= 0 ) { Empty(); return true; }

    Tchar *p = (Tchar*) realloc( m_pstrString, (len + 1) * sizeof(Tchar) );
    if( !p ) return false;

    m_pstrString = p;

    return true;
}

///<param name="zlenSrce">
///  Length of the source string in characters. This length may or may not
///  include the terminating null char. If you include it, it will be converted
///  and copied into the destination string, too. Otherwise not.
///  <br/>
///  If this param is -1, then the source string is imagined to be
///  null terminated and it will be converted including the null terminator.
///</param>
template< class Tchar >
bool miniString<Tchar>::Convert( PWSTR pstrDest, PCSTR pstrSrce, int zlenSrce )
{
    if( !pstrDest || !pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    if( zlenSrce < 0 )
        zlenSrce = (int) ::strlen( pstrSrce ) + 1;
    if( zlenSrce <= 0 ) return true;

    INT ret = MultiByteToWideChar( CP_ACP, 0, pstrSrce, zlenSrce, pstrDest, zlenSrce );
    if( ret == 0 ) return false;

    return true;
}

///<param name="zlenSrce">
///  Length of the source string in characters. This length may or may not
///  include the terminating null char. If you include it, it will be converted
///  and copied into the destination string, too. Otherwise not.
///  <br/>
///  If this param is -1, then the source string is imagined to be
///  null terminated and it will be converted including the null terminator.
///</param>
template< class Tchar >
bool miniString<Tchar>::Convert( PSTR pstrDest, PCWSTR pstrSrce, int zlenSrce )
{
    if( !pstrDest || !pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    if( zlenSrce < 0 )
        zlenSrce = (int) ::wcslen( pstrSrce ) + 1;
    if( zlenSrce <= 0 ) return true;

    INT ret = WideCharToMultiByte( CP_ACP, 0, pstrSrce, zlenSrce, pstrDest, zlenSrce, NULL, NULL );
    if( ret == 0 ) return false;

    return true;
}

template< class Tchar >
bool miniString<Tchar>::Convert( PSTR pstrDest, PCSTR pstrSrce, int zlenSrce )
{
    if( !pstrDest || !pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    memcpy( pstrDest, pstrSrce,
            (zlenSrce < 0 ? ::strlen(pstrSrce) + 1 : zlenSrce) * sizeof(CHAR) );

    return true;
}

template< class Tchar >
bool miniString<Tchar>::Convert( PWSTR pstrDest, PCWSTR pstrSrce, int zlenSrce )
{
    if( !pstrDest || !pstrSrce ) return false;
    if( zlenSrce == 0 ) return true;

    memcpy( pstrDest, pstrSrce,
            (zlenSrce < 0 ? ::wcslen(pstrSrce) + 1 : zlenSrce) * sizeof(WCHAR) );

    return true;
}




}

#endif
