/*
    File:               argonProcessesModules.h
    Description:        Enumeration Processes and its Modules
                        is easily possible using these classes.
                        It will at runtime automtaically use
                        ToolHelp-Api or PSAPI depending on what
                        is present on the current system.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         25.11.2005
    last updated:       02.02.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonProcessesModules_h
#define __argonProcessesModules_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

/*
    Define switches:

    _argon_no_psapi             if defined, PSAPI will not be used at all in here.
    _argon_no_toolhelp32api     if defined, ToolHelp 32 API will not be used at all in here.
*/

#ifndef _argon_no_includes
    #ifndef _argon_no_psapi
    #include <psapi.h>
    #endif
    #ifndef _argon_no_toolhelp32api
    #include <tlhelp32.h>
    #endif

    #include "argonString.h"
    #include "argonArray.h"
#endif

namespace ArgonLibrary
{




/*
    -----------------------------------------------------------------
    Usage:

    In here, you will find the following relevant classes:
        PsApiDll
        ToolHelp32Api
        ProcessesModules

    The first 2 classes are used for dynamically loading the appropriate
    API functions, which are available only on some versions of Windows.

    ProcessesModules is used to have a general API for enumerating Processes
    and their Modules on all Windows versions which will automatically use
    PSAPI or TOOLHELP API depending on which is available.
    You can use the classes
        ProcessesSnapshot
        ModulesSnapshot
    together with ProcessesModules to get snapshot-lists of processes
    and their modules.

    The first 2 Api classes and their related defines will redirect some
    PSAPI and TOOLHELP API function calls to functions defined in here
    which will try to load the real functions dynamically if available
    or fail otherwise.

    Besides that, there are new global functions available:
        GetModuleFileNameExStr


    NOTES:
    When enumerating processes using PSAPI, for some of the processes,
    opening the process to get its HANDLE will be prohibited.
    This is the case when the user process does not have rights to access
    the procress.
    To come around this, you can enable debug privileges "SeDebugPrivilege"
    or SE_DEBUG_NAME using ::AdjustTokenPrivileges() for the *calling*
    process (that is the process who wants to enumerate the processes).
    The calling process will then be seen as a debugger. At least it has
    the privileges of a debugger.
    Then, only the processes with PID 0 (idle process) and 4 (System) will
    be the only ones not accessible (at least on WinXP).
    The ToolHelp API will return names for all processes, but without
    their paths.
    PLUS:
    When setting the calling process to "SeDebugPrivileges" privilege,
    then it is also allowed to terminate almost all processes using
    ::TerminateProcess().
    Unfortunately,
    the privilege API is available only for the WinNT/2k/XP series
    but not for the Win9x/ME series.


    BEWARE:
    The PSAPI has functions where some are available on some systems,
    others not. E.g. there are functions for WinNT/2k/XP, some only
    for Win2k/XP and some only for WinXP. You will have to check yourself,
    which of the functions has been retrieved successfully by class PsApiDll.
*/

#ifndef _argon_no_psapi

#define PSAPIDLL_NAME               "psapi.dll"


#if defined (__cplusplus)
extern "C" {
#endif

typedef BOOL (WINAPI *TfnEnumProcesses)(
    DWORD * lpidProcess,
    DWORD   cb,
    DWORD * cbNeeded
);

BOOL WINAPI argonEnumProcesses(
    DWORD * lpidProcess,
    DWORD   cb,
    DWORD * cbNeeded
);

typedef BOOL (WINAPI *TfnEnumProcessModules)(
    HANDLE hProcess,
    HMODULE *lphModule,
    DWORD cb,
    LPDWORD lpcbNeeded
);

BOOL WINAPI argonEnumProcessModules(
    HANDLE hProcess,
    HMODULE *lphModule,
    DWORD cb,
    LPDWORD lpcbNeeded
);

typedef DWORD (WINAPI *TfnGetModuleBaseNameA)(
    HANDLE hProcess,
    HMODULE hModule,
    LPSTR lpBaseName,
    DWORD nSize
);

DWORD WINAPI argonGetModuleBaseNameA(
    HANDLE hProcess,
    HMODULE hModule,
    LPSTR lpBaseName,
    DWORD nSize
);

typedef DWORD (WINAPI *TfnGetModuleBaseNameW)(
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpBaseName,
    DWORD nSize
);

DWORD WINAPI argonGetModuleBaseNameW(
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpBaseName,
    DWORD nSize
);

typedef DWORD (WINAPI *TfnGetModuleFileNameExA)(
    HANDLE hProcess,
    HMODULE hModule,
    LPSTR lpFilename,
    DWORD nSize
);

DWORD WINAPI argonGetModuleFileNameExA(
    HANDLE hProcess,
    HMODULE hModule,
    LPSTR lpFilename,
    DWORD nSize
);

typedef DWORD (WINAPI *TfnGetModuleFileNameExW)(
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpFilename,
    DWORD nSize
);

DWORD WINAPI argonGetModuleFileNameExW(
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpFilename,
    DWORD nSize
);

typedef BOOL (WINAPI *TfnGetModuleInformation)(
    HANDLE hProcess,
    HMODULE hModule,
    LPMODULEINFO lpmodinfo,
    DWORD cb
);

BOOL WINAPI argonGetModuleInformation(
    HANDLE hProcess,
    HMODULE hModule,
    LPMODULEINFO lpmodinfo,
    DWORD cb
);

typedef BOOL (WINAPI *TfnGetProcessMemoryInfo)(
    HANDLE Process,
    PPROCESS_MEMORY_COUNTERS ppsmemCounters,
    DWORD cb
);

BOOL WINAPI argonGetProcessMemoryInfo(
    HANDLE Process,
    PPROCESS_MEMORY_COUNTERS ppsmemCounters,
    DWORD cb
);

typedef BOOL (WINAPI *TfnGetPerformanceInfo)(
    PPERFORMACE_INFORMATION pPerformanceInformation,
    DWORD cb
);

BOOL WINAPI argonGetPerformanceInfo(
    PPERFORMACE_INFORMATION pPerformanceInformation,
    DWORD cb
);

typedef DWORD (WINAPI *TfnGetProcessImageFileNameA)(
    HANDLE hProcess,
    LPSTR lpImageFileName,
    DWORD nSize
);

DWORD WINAPI argonGetProcessImageFileNameA(
    HANDLE hProcess,
    LPSTR lpImageFileName,
    DWORD nSize
);

typedef DWORD (WINAPI *TfnGetProcessImageFileNameW)(
    HANDLE hProcess,
    LPWSTR lpImageFileName,
    DWORD nSize
);

DWORD WINAPI argonGetProcessImageFileNameW(
    HANDLE hProcess,
    LPWSTR lpImageFileName,
    DWORD nSize
);

#define EnumProcesses               ArgonLibrary::argonEnumProcesses
#define EnumProcessModules          ArgonLibrary::argonEnumProcessModules
#define GetModuleBaseNameA          ArgonLibrary::argonGetModuleBaseNameA
#define GetModuleBaseNameW          ArgonLibrary::argonGetModuleBaseNameW
#define GetModuleFileNameExA        ArgonLibrary::argonGetModuleFileNameExA
#define GetModuleFileNameExW        ArgonLibrary::argonGetModuleFileNameExW
#define GetModuleInformation        ArgonLibrary::argonGetModuleInformation
#define GetProcessMemoryInfo        ArgonLibrary::argonGetProcessMemoryInfo
#define GetPerformanceInfo          ArgonLibrary::argonGetPerformanceInfo
#define GetProcessImageFileNameA    ArgonLibrary::argonGetProcessImageFileNameA
#define GetProcessImageFileNameW    ArgonLibrary::argonGetProcessImageFileNameW

#ifdef UNICODE
    #define argonGetModuleBaseName          argonGetModuleBaseNameW
    #define argonGetModuleFileNameEx        argonGetModuleFileNameExW
    #define argonGetProcessImageFileName    argonGetProcessImageFileNameW
#else
    #define argonGetModuleBaseName          argonGetModuleBaseNameA
    #define argonGetModuleFileNameEx        argonGetModuleFileNameExA
    #define argonGetProcessImageFileName    argonGetProcessImageFileNameA
#endif

#if defined (__cplusplus)
}
#endif

String GetModuleFileNameExStr( HANDLE hProcess, HMODULE hModule );


class PsApiDll
{
    protected :
        class Unloader
        {
            public :
                Unloader() {  }
                ~Unloader() { if( PsApiDll::s_bAutoUnload ) PsApiDll::UnloadAPI(); }
        };

        static Unloader s_Unloader;

    public :
        static HMODULE s_hModule;
        static TfnEnumProcesses            s_pfnEnumProcesses;
        static TfnEnumProcessModules       s_pfnEnumProcessModules;
        static TfnGetModuleBaseNameA       s_pfnGetModuleBaseNameA;
        static TfnGetModuleBaseNameW       s_pfnGetModuleBaseNameW;
        static TfnGetModuleFileNameExA     s_pfnGetModuleFileNameExA;
        static TfnGetModuleFileNameExW     s_pfnGetModuleFileNameExW;
        static TfnGetModuleInformation     s_pfnGetModuleInformation;
        static TfnGetProcessMemoryInfo     s_pfnGetProcessMemoryInfo;
        static TfnGetPerformanceInfo       s_pfnGetPerformanceInfo;
        static TfnGetProcessImageFileNameA s_pfnGetProcessImageFileNameA;
        static TfnGetProcessImageFileNameW s_pfnGetProcessImageFileNameW;
        static bool s_bAutoUnload;  // true = automatically unloads the Dll when global objects are destructed.

        PsApiDll();

        static bool PsApiExists();
        static bool hasGetPerformanceInfo();

        static void LoadAPI();
        static void UnloadAPI();

    protected :
        static void Nullify();
};

#endif




#ifndef _argon_no_toolhelp32api

#define TOOLHELP32MODULE_NAME       "kernel32.dll"


#if defined (__cplusplus)
extern "C" {
#endif

typedef HANDLE (WINAPI *TfnCreateToolhelp32Snapshot)(
    DWORD dwFlags,
    DWORD th32ProcessID
);

HANDLE WINAPI argonCreateToolhelp32Snapshot(
    DWORD dwFlags,
    DWORD th32ProcessID
);

typedef BOOL (WINAPI *TfnHeap32ListFirst)(
    HANDLE hSnapshot,
    LPHEAPLIST32 lphl
);

BOOL WINAPI argonHeap32ListFirst(
    HANDLE hSnapshot,
    LPHEAPLIST32 lphl
);

typedef BOOL (WINAPI *TfnHeap32ListNext)(
    HANDLE hSnapshot,
    LPHEAPLIST32 lphl
);

BOOL WINAPI argonHeap32ListNext(
    HANDLE hSnapshot,
    LPHEAPLIST32 lphl
);

typedef BOOL (WINAPI *TfnHeap32First)(
    LPHEAPENTRY32 lphe,
    DWORD th32ProcessID,
    ULONG_PTR th32HeapID
);

BOOL WINAPI argonHeap32First(
    LPHEAPENTRY32 lphe,
    DWORD th32ProcessID,
    ULONG_PTR th32HeapID
);

typedef BOOL (WINAPI *TfnHeap32Next)(
    LPHEAPENTRY32 lphe
);

BOOL WINAPI argonHeap32Next(
    LPHEAPENTRY32 lphe
);

typedef BOOL (WINAPI *TfnToolhelp32ReadProcessMemory)(
    DWORD   th32ProcessID,
    LPCVOID lpBaseAddress,
    LPVOID  lpBuffer,
    SIZE_T  cbRead,
    SIZE_T *lpNumberOfBytesRead
);

BOOL WINAPI argonToolhelp32ReadProcessMemory(
    DWORD   th32ProcessID,
    LPCVOID lpBaseAddress,
    LPVOID  lpBuffer,
    SIZE_T  cbRead,
    SIZE_T *lpNumberOfBytesRead
);

typedef BOOL (WINAPI *TfnProcess32FirstW)(
    HANDLE hSnapshot,
    LPPROCESSENTRY32W lppe
);

BOOL WINAPI argonProcess32FirstW(
    HANDLE hSnapshot,
    LPPROCESSENTRY32W lppe
);

typedef BOOL (WINAPI *TfnProcess32NextW)(
    HANDLE hSnapshot,
    LPPROCESSENTRY32W lppe
);

BOOL WINAPI argonProcess32NextW(
    HANDLE hSnapshot,
    LPPROCESSENTRY32W lppe
);

typedef BOOL (WINAPI *TfnProcess32First)(
    HANDLE hSnapshot,
    tagPROCESSENTRY32 *lppe  // LPPROCESSENTRY32 lppe
);

BOOL WINAPI argonProcess32First(
    HANDLE hSnapshot,
    tagPROCESSENTRY32 *lppe  // LPPROCESSENTRY32 lppe
);

typedef BOOL (WINAPI *TfnProcess32Next)(
    HANDLE hSnapshot,
    tagPROCESSENTRY32 *lppe  // LPPROCESSENTRY32 lppe
);

BOOL WINAPI argonProcess32Next(
    HANDLE hSnapshot,
    tagPROCESSENTRY32 *lppe  // LPPROCESSENTRY32 lppe
);

typedef BOOL (WINAPI *TfnThread32First)(
    HANDLE hSnapshot,
    LPTHREADENTRY32 lpte
);

BOOL WINAPI argonThread32First(
    HANDLE hSnapshot,
    LPTHREADENTRY32 lpte
);

typedef BOOL (WINAPI *TfnThread32Next)(
    HANDLE hSnapshot,
    LPTHREADENTRY32 lpte
);

BOOL WINAPI argonThread32Next(
    HANDLE hSnapshot,
    LPTHREADENTRY32 lpte
);

typedef BOOL (WINAPI *TfnModule32FirstW)(
    HANDLE hSnapshot,
    LPMODULEENTRY32W lpme
);

BOOL WINAPI argonModule32FirstW(
    HANDLE hSnapshot,
    LPMODULEENTRY32W lpme
);

typedef BOOL (WINAPI *TfnModule32NextW)(
    HANDLE hSnapshot,
    LPMODULEENTRY32W lpme
);

BOOL WINAPI argonModule32NextW(
    HANDLE hSnapshot,
    LPMODULEENTRY32W lpme
);

typedef BOOL (WINAPI *TfnModule32First)(
    HANDLE hSnapshot,
    tagMODULEENTRY32 *lpme  // LPMODULEENTRY32 lpme
);

BOOL WINAPI argonModule32First(
    HANDLE hSnapshot,
    tagMODULEENTRY32 *lpme  // LPMODULEENTRY32 lpme
);

typedef BOOL (WINAPI *TfnModule32Next)(
    HANDLE hSnapshot,
    tagMODULEENTRY32 *lpme  // LPMODULEENTRY32 lpme
);

BOOL WINAPI argonModule32Next(
    HANDLE hSnapshot,
    tagMODULEENTRY32 *lpme  // LPMODULEENTRY32 lpme
);

#define CreateToolhelp32Snapshot    ArgonLibrary::argonCreateToolhelp32Snapshot
#define Heap32ListFirst             ArgonLibrary::argonHeap32ListFirst
#define Heap32ListNext              ArgonLibrary::argonHeap32ListNext
#define Heap32First                 ArgonLibrary::argonHeap32First
#define Heap32Next                  ArgonLibrary::argonHeap32Next
#define Toolhelp32ReadProcessMemory ArgonLibrary::argonToolhelp32ReadProcessMemory
#define Process32FirstW             ArgonLibrary::argonProcess32FirstW
#define Process32NextW              ArgonLibrary::argonProcess32NextW
#undef  Process32First
#undef  Process32Next
#define Process32First              ArgonLibrary::argonProcess32First
#define Process32Next               ArgonLibrary::argonProcess32Next
#define Thread32First               ArgonLibrary::argonThread32First
#define Thread32Next                ArgonLibrary::argonThread32Next
#define Module32FirstW              ArgonLibrary::argonModule32FirstW
#define Module32NextW               ArgonLibrary::argonModule32NextW
#undef  Module32First
#undef  Module32Next
#define Module32First               ArgonLibrary::argonModule32First
#define Module32Next                ArgonLibrary::argonModule32Next

#ifdef UNICODE
    #define argonProcess32First     argonProcess32FirstW
    #define argonProcess32Next      argonProcess32NextW
    #define argonModule32First      argonModule32FirstW
    #define argonModule32Next       argonModule32NextW
#endif

#if defined (__cplusplus)
}
#endif


class ToolHelp32Api
{
    protected :
        class Unloader
        {
            public :
                Unloader() {  }
                ~Unloader() { if( ToolHelp32Api::s_bAutoUnload ) ToolHelp32Api::UnloadAPI(); }
        };

        static Unloader s_Unloader;

    public :
        static HMODULE s_hModule;
        static TfnCreateToolhelp32Snapshot    s_pfnCreateToolhelp32Snapshot;
        static TfnHeap32ListFirst             s_pfnHeap32ListFirst;
        static TfnHeap32ListNext              s_pfnHeap32ListNext;
        static TfnHeap32First                 s_pfnHeap32First;
        static TfnHeap32Next                  s_pfnHeap32Next;
        static TfnToolhelp32ReadProcessMemory s_pfnToolhelp32ReadProcessMemory;
        static TfnProcess32FirstW             s_pfnProcess32FirstW;
        static TfnProcess32NextW              s_pfnProcess32NextW;
        static TfnProcess32First              s_pfnProcess32First;
        static TfnProcess32Next               s_pfnProcess32Next;
        static TfnThread32First               s_pfnThread32First;
        static TfnThread32Next                s_pfnThread32Next;
        static TfnModule32FirstW              s_pfnModule32FirstW;
        static TfnModule32NextW               s_pfnModule32NextW;
        static TfnModule32First               s_pfnModule32First;
        static TfnModule32Next                s_pfnModule32Next;
        static bool s_bAutoUnload;  // true = automatically unloads the Dll when global objects are destructed.

        ToolHelp32Api();

        static bool ToolHelp32ApiExists();

        static void LoadAPI();
        static void UnloadAPI();

    protected :
        static void Nullify();
};

#endif




class ProcessesSnapshot
{
    public :
        struct Info
        {
            DWORD   m_dwPid;
            TCHAR   m_szName    [_MAX_PATH_ + 1];
            TCHAR   m_szFilepath[_MAX_PATH_ + 1];
        };

        Array<Info, 0>  m_Snapshot;

    public :
        ProcessesSnapshot() { m_Snapshot.SetBlockSize(128); }

        void Empty() { m_Snapshot.Empty(); }
        bool isEmpty() const { return m_Snapshot.GetSize() == 0; }

        Info    *getPtr  () { return m_Snapshot.Ptr(); }
        _size_t_ getCount() const { return m_Snapshot.GetSize(); }

        Info *getAt( _idx_ idx ) { return m_Snapshot.PtrAt( idx ); }
        Info *findByPid( DWORD pid );
};

class ModulesSnapshot
{
    public :
        struct Info
        {
            HMODULE     m_hModule;
            void       *m_pBaseAddr;
            _size_t_    m_nBaseSize;
            TCHAR       m_szFilename[_MAX_PATH_ + 1];
        };

        DWORD           m_dwPid;    // the process ID for which the list has been retrieved
        Array<Info, 0>  m_Snapshot;

    public :
        ModulesSnapshot() { m_dwPid = 0; m_Snapshot.SetBlockSize(128); }

        void Empty() { m_Snapshot.Empty(); m_dwPid = 0; }
        bool isEmpty() const { return m_Snapshot.GetSize() == 0; }

        Info    *getPtr () { return m_Snapshot.Ptr(); }
        _size_t_ getCount() const { return m_Snapshot.GetSize(); }

        Info *getAt( _idx_ idx ) { return m_Snapshot.PtrAt( idx ); }
        Info *findByHModule( HMODULE hMod );
};

class ProcessesModules
{
    protected :
        bool    m_bUsePsApi;
        bool    m_bUseToolHelpApi;

    public :
        ProcessesModules()
            { m_bUsePsApi = m_bUseToolHelpApi = false; m_pstrUnknownName = _T("--unknown--"); }
        ~ProcessesModules() {  }

    public :
            ///<summary>
            ///  Call <b>Init()</b> first to initialize the API and select if
            ///  you prefer the TOOLHELP API or the PSAPI.
            ///</summary>
            ///<param name="bPsApi_preferred"> This parameter is relevant on system
            ///  which have both, the PSAPI as well as the TOOLHELP API. <br/>
            ///  By setting this to true, the method will try to use the PSAPI
            ///  if available. Otherwise it will use the TOOLHELP API. <br/>
            ///  By setting this to fase, first the TOOLHELP API will tried to be
            ///  loaded and only if that fails, the PSAPI will be used instead. </param>
            ///<returns> True on success, false on error (if no API could be retrieved). </returns>
        bool Init( bool bPsApi_preferred = true );

            ///<summary>
            ///  <b>enumProcesses()</b> returns a snapshot of all the processes
            ///  currently running on the system.
            ///</summary>
        bool enumProcesses( ProcessesSnapshot& snapshot );
            ///<summary>
            ///  <b>enumProcessModules()</b> returns a snapshot of all the modules
            ///  currently loaded in the process identified by the specified
            ///  process-ID.
            ///</summary>
            ///<param name="maxModules"> maximum number of modules to retrieve.
            ///  If negative, there is no upper limit. </param>
        bool enumProcessModules( ModulesSnapshot& snapshot, DWORD pid, int maxModules = -1 );

    protected :
        bool _enumProcessesPsApi( ProcessesSnapshot& snapshot );
        bool _enumProcessesToolHelpApi( ProcessesSnapshot& snapshot );
        bool _enumProcessModulesPsApi( ModulesSnapshot& snapshot, DWORD pid, int maxModules );
        bool _enumProcessModulesToolHelpApi( ModulesSnapshot& snapshot, DWORD pid, int maxModules );
        bool _getProcessModuleFilepathToolHelpApi( PTSTR pszFilepath, DWORD pid, _size_t_ cFilepath );

    public :
        PCTSTR  m_pstrUnknownName;  // name to be used for a process or module when it cannot be retrieved (*must* always be a valid string)
};




}

#endif
