/*
    File:               argonResources.h
    Description:        Simple classes for easy CRT and Windows resource handling.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         25.08.2005
    last updated:       01.10.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonResources_h
#define __argonResources_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




/*
    The classes in here can be used to automatically release a resource
    or automatically undo an operation.
    Choose the appropriate class for your resource and call
    #.releaseResource()# or let the destructor (dtor) call it for you.
    The resource will be released then.

    There are classes or templates for:
        delete and delete[] operators
        C runtime resources
        Window objects
*/



//  delete and delete[] operators

///<summary>
///  This class calls the delete operator. (non-array version). <br/>
///  In addition, this class can act like a smart pointer class regarding
///  the most basic smart pointer functionalities (well the pointer access).
///</summary>
template< class T >
class deleteOp
{
    public :
        T *m_ptr;

    public :
        deleteOp() { m_ptr = NULL; }
        deleteOp( T *ptr ) { m_ptr = ptr; }
        ~deleteOp() { releaseResource(); }

        deleteOp& operator =( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        deleteOp& set( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        T        *get() { return m_ptr; }
        T        *detach() { T *p = m_ptr; m_ptr = NULL; return p; }
        void      releaseResource() { if( m_ptr ) { delete m_ptr; m_ptr = NULL; } }

        operator T       *()       { return m_ptr; }
        operator const T *() const { return m_ptr; }
        T&       operator *()       { return *m_ptr; }
        const T& operator *() const { return *m_ptr; }
        T       *operator ->()       { return m_ptr; }
        const T *operator ->() const { return m_ptr; }

    protected :
        deleteOp( const deleteOp& src );
        deleteOp& operator =( const deleteOp& src );
};

///<summary>
///  This class calls the delete[] operator. (array version). <br/>
///  In addition, this class can act like a smart pointer class regarding
///  the most basic smart pointer functionalities (well the pointer access).
///</summary>
template< class T >     // T is the type of one item in the array
class deleteAr
{
    public :
        T *m_ptr;

    public :
        deleteAr() { m_ptr = NULL; }
        deleteAr( T *ptr ) { m_ptr = ptr; }
        ~deleteAr() { releaseResource(); }

        deleteAr& operator =( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        deleteAr& set( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        T        *get() { return m_ptr; }
        T        *detach() { T *p = m_ptr; m_ptr = NULL; return p; }
        void      releaseResource() { if( m_ptr ) { delete[] m_ptr; m_ptr = NULL; } }

        operator T       *()       { return m_ptr; }
        operator const T *() const { return m_ptr; }
        T&       operator *()       { return *m_ptr; }
        const T& operator *() const { return *m_ptr; }
        T       *operator ->()       { return m_ptr; }
        const T *operator ->() const { return m_ptr; }

    protected :
        deleteAr( const deleteAr& src );
        deleteAr& operator =( const deleteAr& src );
};



//  C runtime resources

///<summary>
///  Class <b>freeMalloc</b> allows automatic freeing of memory using the
///  CRT call <b>free()</b>. <br/>
///  You can assign or get the contained pointer to the memory.
///  So in that way, this class acts like a pointer.
///  When assigning a new pointer, the previous memory (if any)
///  will first be freed automatically.
///</summary>
template< class T >
class freeMalloc
{
    public :
        T *m_ptr;

    public :
        freeMalloc() { m_ptr = NULL; }
        freeMalloc( T *ptr ) { m_ptr = ptr; }
        ~freeMalloc() { releaseResource(); }

        freeMalloc& operator =( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        freeMalloc& set( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        T          *get() const { return m_ptr; }
        T          *detach() { T *p = m_ptr; m_ptr = NULL; return p; }
        void        releaseResource() { if( m_ptr ) { ::free( m_ptr ); m_ptr = NULL; } }

        operator T       *()       { return m_ptr; }
        operator const T *() const { return m_ptr; }

    protected :
        freeMalloc( const freeMalloc& src );
        freeMalloc& operator =( const freeMalloc& src );
};

///<summary>
///  Class <b>fcloseFile</b> allows automatic closing of a CRT file or stream
///  using the CRT call <b>fclose()</b>. <br/>
///  You can assign or get the contained file pointer.
///  So in that way, this class acts like a FILE* pointer.
///  When assigning a new pointer, the previous file (if any)
///  will first be closed automatically.
///</summary>
class fcloseFile
{
    public :
        FILE *m_stream;

    public :
        fcloseFile() { m_stream = NULL; }
        fcloseFile( FILE *stream ) { m_stream = stream; }
        ~fcloseFile() { releaseResource(); }

        fcloseFile& operator =( FILE *stream ) { releaseResource(); m_stream = stream; return *this; }
        fcloseFile& set( FILE *stream ) { releaseResource(); m_stream = stream; return *this; }
        FILE       *get() const { return m_stream; }
        FILE       *detach() { FILE *f = m_stream; m_stream = NULL; return f; }
        void        releaseResource() { if( m_stream ) { ::fclose( m_stream ); m_stream = NULL; } }

        operator FILE *() { return m_stream; }

    protected :
        fcloseFile( const fcloseFile& src );
        fcloseFile& operator =( const fcloseFile& src );
};



//  Windows objects

///<summary>
///  Class <b>delDC</b> allows automatic deletion of a DC
///  using <b>DeleteDC()</b>. <br/>
///  You can assign or get the contained HDC handle.
///  When assigning a new HDC, the previous DC (if any)
///  will first be deleted automatically.
///</summary>
class delDC
{
    public :
        HDC m_hdc;

    public :
        delDC() { m_hdc = NULL; }
        delDC( HDC hdc ) { m_hdc = hdc; }
        ~delDC() { releaseResource(); }

        delDC& operator =( HDC hdc ) { releaseResource(); m_hdc = hdc; return *this; }
        delDC& set( HDC hdc ) { releaseResource(); m_hdc = hdc; return *this; }
        HDC    get() const { return m_hdc; }
        HDC    detach() { HDC hdc = m_hdc; m_hdc = NULL; return hdc; }
        void   releaseResource() { if( m_hdc ) { ::DeleteDC( m_hdc ); m_hdc = NULL; } }

        operator HDC() { return m_hdc; }

    protected :
        delDC( const delDC& src );
        delDC& operator =( const delDC& src );
};

///<summary>
///  Class <b>releaseDC</b> allows automatic releasing of a DC
///  using <b>ReleaseDC()</b>. Usually you will use it to release
///  a DC that has been retrieved by calls like <b>GetDC()</b>
///  or <b>GetWindowDC()</b>. <br/>
///  You can set or get the contained HDC handle.
///  When setting a new HDC, the previous DC (if any)
///  will first be released automatically.
///</summary>
class releaseDC
{
    public :
        HWND m_hwnd;
        HDC  m_hdc;

    public :
        releaseDC() { m_hwnd = NULL; m_hdc = NULL; }
        releaseDC( HWND hwnd, HDC hdc ) { m_hwnd = hwnd; m_hdc = hdc; }
        ~releaseDC() { releaseResource(); }

        releaseDC& set( HWND hwnd, HDC hdc ) { releaseResource(); m_hwnd = hwnd; m_hdc = hdc; return *this; }
        HDC        get() const { return m_hdc; }
        HWND       getHwnd() const { return m_hwnd; }
        void       detach() { m_hwnd = NULL; m_hdc = NULL; }
        void       releaseResource() { if( m_hdc ) { ::ReleaseDC( m_hwnd, m_hdc ); m_hwnd = NULL; m_hdc = NULL; } }

        operator HDC() { return m_hdc; }

    protected :
        releaseDC( const releaseDC& src );
        releaseDC& operator =( const releaseDC& src );
};

///<summary>
///  Class <b>releasePaintDC</b> allows automatic ending of a DC
///  used in a WM_PAINT message using <b>EndPaint()</b>.
///  Usually you will use it to end a DC that has been retrieved
///  by a call to <b>BeginPaint()</b>. <br/>
///  In addition, this class also allows automatic calling of
///  <b>BeginPaint()</b> (in the ctor) as well as <b>EndPaint()</b>
///  (in the dtor) by using the ctor with only the HWND parameter
///  to be specified.
///</summary>
class releasePaintDC
{
    public :
        HWND        m_hwnd;
        HDC         m_hdc;
        PAINTSTRUCT m_paintstruct;  // the content data is valid *only* if #m_hdc# is not NULL

    public :
            ///<summary> Call the ctor with only the HWND parameter to automatically perform
            ///  the <b>BeginPaint()</b> call for you. </summary>
        releasePaintDC( HWND hwnd )
            { m_hwnd = hwnd; m_hdc = hwnd ? ::BeginPaint( hwnd, &m_paintstruct ) : NULL; }
        releasePaintDC( HWND hwnd, HDC hdc, const PAINTSTRUCT& ps ) { m_hwnd = hwnd; m_hdc = hdc; m_paintstruct = ps; }
        ~releasePaintDC() { releaseResource(); }

        releasePaintDC&    set( HWND hwnd, HDC hdc, const PAINTSTRUCT& ps ) { releaseResource(); m_hwnd = hwnd; m_hdc = hdc; m_paintstruct = ps; return *this; }
        HDC                get() const { return m_hdc; }
        HWND               getHwnd() const { return m_hwnd; }
        const PAINTSTRUCT& getPaintStruct() const { return m_paintstruct; }
        void               releaseResource() { if( m_hdc && m_hwnd ) { ::EndPaint( m_hwnd, &m_paintstruct ); m_hwnd = NULL; m_hdc = NULL; } }

        operator HDC() { return m_hdc; }

    protected :
        releasePaintDC( const releasePaintDC& src );
        releasePaintDC& operator =( const releasePaintDC& src );
};

///<summary>
///  <b>memBitmapDC</b> will automatically create a compatible memory DC and
///  a compatible bitmap for you and selects the bitmap into the new dc. <br/>
///  On <b>releaseResource()</b> (or in the dtor), they will be deleted automatically.
///  This is useful when you need a temporary or memory bitmap to draw into.
///</summary>
class memBitmapDC
{
    public :
        HDC     m_hdc;
        HBITMAP m_hbmp;
        HGDIOBJ m_oldbmp;       // handle to the old bitmap (before #m_hbmp# has been selected into #m_hdc#)

    public :
        memBitmapDC() { m_hdc = NULL; m_hbmp = NULL; m_oldbmp = NULL; }
        memBitmapDC( HDC hdcRef, int widthBmp, int heightBmp ) { m_hdc = NULL; m_hbmp = NULL; create( hdcRef, widthBmp, heightBmp ); }
        memBitmapDC( HDC hdcRef, const SIZE& sizBmp ) { m_hdc = NULL; m_hbmp = NULL; create( hdcRef, sizBmp ); }
        ~memBitmapDC() { releaseResource(); }

        memBitmapDC& create( HDC hdcRef, const SIZE& sizBmp ) { return create( hdcRef, (int) sizBmp.cx, (int) sizBmp.cy ); }
        memBitmapDC& create( HDC hdcRef, int widthBmp, int heightBmp )
            { releaseResource();
              m_hdc = hdcRef ? CreateCompatibleDC(hdcRef) : NULL;
              m_hbmp = m_hdc && widthBmp > 0 && heightBmp > 0 ? CreateCompatibleBitmap(hdcRef, widthBmp, heightBmp) : NULL;
              m_oldbmp = m_hbmp && m_hdc ? SelectObject(m_hdc, m_hbmp) : NULL;
              return *this; }

        HDC     getDC()        const { return m_hdc;    }
        HBITMAP getBitmap()    const { return m_hbmp;   }
        HGDIOBJ getOldBitmap() const { return m_oldbmp; }

        void releaseResource()
            { if( m_hbmp && m_hdc ) SelectObject( m_hdc, m_oldbmp );
              if( m_hbmp ) { DeleteObject( m_hbmp ); m_hbmp = NULL; }
              if( m_hdc ) { DeleteDC( m_hdc ); m_hdc = NULL; } }

        operator HDC() { return m_hdc; }

    protected :
        memBitmapDC( const memBitmapDC& src );
        memBitmapDC& operator =( const memBitmapDC& src );
};

class saveDC
{
    public :
        HDC m_hdc;
        int m_nSaved;

    public :
        saveDC() { m_hdc = NULL; m_nSaved = 0; }
        saveDC( HDC hdc ) { m_hdc = hdc; m_nSaved = hdc ? SaveDC(hdc) : 0; }
        ~saveDC() { releaseResource(); }

        saveDC& save( HDC hdc ) { releaseResource(); m_hdc = hdc; m_nSaved = hdc ? SaveDC(hdc) : 0; return *this; }
        void    releaseResource() { if( m_hdc ) { RestoreDC( m_hdc, m_nSaved ); m_hdc = NULL; } }
        HDC     getDC() const { return m_hdc; }

    protected :
        saveDC( const saveDC& src );
        saveDC& operator =( const saveDC& src );
};

///<summary>
///  Class <b>selectObj</b> allows automatic restoring of an object
///  that has been selected into a DC. <br/>
///  In addition, this class allows selecting an object into a DC
///  (in the ctor by setting parameter <b>bSelectObject</b> to true)
///  and having it restored automatically later on (in the dtor
///  or by a call to <b>releaseResource()</b>).
///</summary>
template< class T_HGDIOBJ >     // for T_HGDIOBJ specify HBITMAP, HPALETTE, HPEN, HBRUSH, HFONT, or HGDIOBJ in general.
class selectObj
{
    public :
        T_HGDIOBJ m_oldHandle;
        HDC       m_hdc;

    public :
        selectObj() { m_oldHandle = NULL; m_hdc = NULL; }
        selectObj( HDC hdc, T_HGDIOBJ hObjToSelect_or_ForAutoRelease, bool bSelectObject )
            { m_hdc = hdc; m_oldHandle = hdc ? ( bSelectObject ? (T_HGDIOBJ) ::SelectObject(hdc, (HGDIOBJ) hObjToSelect_or_ForAutoRelease) : hObjToSelect_or_ForAutoRelease ) : NULL; }
        ~selectObj() { releaseResource(); }

        selectObj& set( HDC hdc, T_HGDIOBJ hObjForAutoRelease ) { releaseResource(); m_oldHandle = hObjForAutoRelease; m_hdc = hdc; return *this; }
        void       selectObject( T_HGDIOBJ hObjToSelect ) { selectObject( m_hdc, hObjToSelect ); }
            ///<summary> <b>selectObject()</b> selects another object into the DC. If the new HDC is
            ///  different than the previous one, the old object will be restored first. </summary>
        void       selectObject( HDC hdc, T_HGDIOBJ hObjToSelect )
            { if( hdc && hdc != m_hdc ) releaseResource(); if( !m_hdc ) m_hdc = hdc; if( hdc && hdc == m_hdc ) { T_HGDIOBJ oldH = (T_HGDIOBJ) ::SelectObject( hdc, (HGDIOBJ) hObjToSelect ); if( !m_oldHandle ) m_oldHandle = oldH; } }
        void       releaseResource() { if( m_oldHandle && m_hdc ) { ::SelectObject( m_hdc, (HGDIOBJ) m_oldHandle ); m_oldHandle = NULL; m_hdc = NULL; } }

        HGDIOBJ getOldHandle() const { return m_oldHandle; }
        HDC     getDC()        const { return m_hdc;       }

    protected :
        selectObj( const selectObj& src );
        selectObj& operator =( const selectObj& src );
};

///<summary>
///  Class <b>delObj</b> allows automatic deletion of a gdi object
///  using <b>DeleteObject()</b> (in the dtor or in <b>releaseResource()</b>). <br/>
///  You can assign or get the gdi object handle.
///  So in that way, this class acts like a HGDIOBJ handle.
///  When assigning a new handle, the previous gdi object (if any)
///  will first be deleted automatically.
///</summary>
template< class T_HGDIOBJ >
class delObj
{
    public :
        T_HGDIOBJ m_handle;

    public :
        delObj() { m_handle = NULL; }
        delObj( T_HGDIOBJ h ) { m_handle = h; }
        ~delObj() { releaseResource(); }

        delObj&   operator =( T_HGDIOBJ h ) { releaseResource(); m_handle = h; return *this; }
        delObj&   set( T_HGDIOBJ h ) { releaseResource(); m_handle = h; return *this; }
        T_HGDIOBJ get() const { return m_handle; }
        T_HGDIOBJ detach() { T_HGDIOBJ h = m_handle; m_handle = NULL; return h; }
        void      releaseResource() { if( m_handle ) { ::DeleteObject( m_handle ); m_handle = NULL; } }

        operator T_HGDIOBJ() { return m_handle; }

    protected :
        delObj( const delObj& src );
        delObj& operator =( const delObj& src );
};

///<summary>
///  Class <b>destroyIcon</b> allows automatic destruction of an icon handle
///  using <b>DestroyIcon()</b> (in the dtor or in <b>releaseResource()</b>). <br/>
///  You can assign or get the icon handle.
///  So in that way, this class acts like a HICON handle.
///  When assigning a new handle, the previous icon handle (if any)
///  will first be destroyed automatically.
///</summary>
class destroyIcon
{
    public :
        HICON m_hicon;

    public :
        destroyIcon() { m_hicon = NULL; }
        destroyIcon( HICON hicon ) { m_hicon = hicon; }
        ~destroyIcon() { releaseResource(); }

        destroyIcon& operator =( HICON hicon ) { releaseResource(); m_hicon = hicon; return *this; }
        destroyIcon& set( HICON hicon ) { releaseResource(); m_hicon = hicon; return *this; }
        HICON        get() const { return m_hicon; }
        HICON        detach() { HICON hicon = m_hicon; m_hicon = NULL; return hicon; }
        void         releaseResource() { if( m_hicon ) { ::DestroyIcon( m_hicon ); m_hicon = NULL; } }

        operator HICON() { return m_hicon; }

    protected :
        destroyIcon( const destroyIcon& src );
        destroyIcon& operator =( const destroyIcon& src );
};

#ifndef _WIN32_WCE

///<summary>
///  Class <b>destroyCursor</b> allows automatic destruction of a cursor handle
///  using <b>DestroyCursor()</b> (in the dtor or in <b>releaseResource()</b>). <br/>
///  You can assign or get the cursor handle.
///  So in that way, this class acts like a HCURSOR handle.
///  When assigning a new handle, the previous cursor handle (if any)
///  will first be destroyed automatically.
///</summary>
class destroyCursor
{
    public :
        HCURSOR m_hcursor;

    public :
        destroyCursor() { m_hcursor = NULL; }
        destroyCursor( HCURSOR hcrsr ) { m_hcursor = hcrsr; }
        ~destroyCursor() { releaseResource(); }

        destroyCursor& operator =( HCURSOR hcrsr ) { releaseResource(); m_hcursor = hcrsr; return *this; }
        destroyCursor& set( HCURSOR hcrsr ) { releaseResource(); m_hcursor = hcrsr; return *this; }
        HCURSOR        get() const { return m_hcursor; }
        HCURSOR        detach() { HCURSOR hcrsr = m_hcursor; m_hcursor = NULL; return hcrsr; }
        void           releaseResource() { if( m_hcursor ) { ::DestroyCursor( m_hcursor ); m_hcursor = NULL; } }

        operator HCURSOR() { return m_hcursor; }

    protected :
        destroyCursor( const destroyCursor& src );
        destroyCursor& operator =( const destroyCursor& src );
};

#endif

///<summary>
///  Class <b>closeH</b> allows automatic closing of a HANDLE
///  using <b>CloseHandle()</b> (in the dtor or in <b>releaseResource()</b>). <br/>
///  You can assign or get the handle.
///  So in that way, this class acts like a HANDLE handle.
///  When assigning a new handle, the previous handle (if any)
///  will first be closed automatically.
///</summary>
class closeH
{
    public :
        HANDLE m_handle;

    public :
        closeH() { m_handle = NULL; }
        closeH( HANDLE h ) { m_handle = h; }
        ~closeH() { releaseResource(); }

        closeH& operator =( HANDLE h ) { releaseResource(); m_handle = h; return *this; }
        closeH& set( HANDLE h ) { releaseResource(); m_handle = h; return *this; }
        HANDLE  get() const { return m_handle; }
        HANDLE  detach() { HANDLE h = m_handle; m_handle = NULL; return h; }
        void    releaseResource() { if( m_handle ) { ::CloseHandle( m_handle ); m_handle = NULL; } }

        operator HANDLE() { return m_handle; }

    protected :
        closeH( const closeH& src );
        closeH& operator =( const closeH& src );
};

///<summary>
///  Class <b>unmapVof</b> allows automatic unmapping of a view of file
///  using <b>UnmapViewOfFile()</b> (in the dtor or in <b>releaseResource()</b>)
///  as a <b>const void*</b> pointer. <br/>
///  You can assign or get the pointer.
///  So in that way, this class acts like a <b>const void*</b> pointer.
///  When assigning a new pointer, the previous pointer (if any)
///  will first be unmapped automatically.
///</summary>
class unmapVof
{
    public :
        const void *m_ptr;

    public :
        unmapVof() { m_ptr = NULL; }
        unmapVof( const void *ptr ) { m_ptr = ptr; }
        ~unmapVof() { releaseResource(); }

        unmapVof&   operator =( const void *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        unmapVof&   set( const void *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        const void *get() const { return m_ptr; }
        const void *detach() { const void *p = m_ptr; m_ptr = NULL; return p; }
        void        releaseResource() { if( m_ptr ) { ::UnmapViewOfFile( m_ptr ); m_ptr = NULL; } }

        operator const void *() const { return m_ptr; }

    protected :
        unmapVof( const unmapVof& src );
        unmapVof& operator =( const unmapVof& src );
};

///<summary>
///  Class <b>freelibH</b> allows automatic freeing of a HMODULE library
///  using <b>FreeLibrary()</b> (in the dtor or in <b>releaseResource()</b>). <br/>
///  You can assign or get the handle.
///  So in that way, this class acts like a HMODULE handle.
///  When assigning a new handle, the previous handle (if any)
///  will first be freed automatically.
///</summary>
class freelibH
{
    public :
        HMODULE m_hModule;

    public :
        freelibH() { m_hModule = NULL; }
        freelibH( HMODULE hmod ) { m_hModule = hmod; }
        ~freelibH() { releaseResource(); }

        freelibH& operator =( HMODULE hmod ) { releaseResource(); m_hModule = hmod; return *this; }
        freelibH& set( HMODULE hmod ) { releaseResource(); m_hModule = hmod; return *this; }
        HMODULE   get() const { return m_hModule; }
        HMODULE   detach() { HMODULE hmod = m_hModule; m_hModule = NULL; return hmod; }
        void      releaseResource() { if( m_hModule ) { ::FreeLibrary( m_hModule ); m_hModule = NULL; } }

        operator HMODULE() { return m_hModule; }

    protected :
        freelibH( const freelibH& src );
        freelibH& operator =( const freelibH& src );
};



//  COM objects

///<summary>
///  This class calls the <b>Release()</b> method of the contained object. <br/>
///  In addition, this class can act like a smart pointer class regarding
///  the most basic smart pointer functionalities (well the pointer access). <br/>
///  Note: <br/>
///  It is not really necessary that the contained object is a COM object.
///  It is totally sufficient, if the contained object just has a method
///      <b>void Release()</b>
///  that releases itself. <br/>
///  IMPORTANT: <br/>
///  As the name of the template says, it releases an object. So it does
///  *never* call <b>AddRef()</b> of the contained object.
///</summary>
template< class T >
class releaseCom
{
    public :
        T *m_ptr;

    public :
        releaseCom() { m_ptr = NULL; }
        releaseCom( T *ptr ) { m_ptr = ptr; }
        ~releaseCom() { releaseResource(); }

        releaseCom& operator =( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        releaseCom& set( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        T          *get() { return m_ptr; }
        T          *detach() { T *p = m_ptr; m_ptr = NULL; return p; }
        void        releaseResource() { if( m_ptr ) { m_ptr->Release(); m_ptr = NULL; } }

        operator T       *()       { return m_ptr; }
        operator const T *() const { return m_ptr; }
        T&       operator *()       { return *m_ptr; }
        const T& operator *() const { return *m_ptr; }
        T       *operator ->()       { return m_ptr; }
        const T *operator ->() const { return m_ptr; }
            ///<summary> The & operator (addr-of) will normally be used to pass the pointer as parameter
            ///  to a function that returns the new COM object. So we expect this behaviour as the
            ///  only desired one and therefore first release any previous COM object, if any. </summary>
        T **operator &() { releaseResource(); return &m_ptr; }

    protected :
        releaseCom( const releaseCom& src );
        releaseCom& operator =( const releaseCom& src );
};



//  misc ...

///<summary>
///  This class calls the <b>Empty()</b> method of the contained object. <br/>
///  The class <b>T</b> must have a method <b>Empty()</b> that has no parameters. <br/>
///  Besides the usual releasing of the resource, it has a <b>detach()</b> method
///  that detaches the contained pointer *without* calling its <b>Empty()</b> method.
///</summary>
template< class T >
class autoEmpty
{
    public :
        T *m_ptr;

    public :
        autoEmpty() { m_ptr = NULL; }
        autoEmpty( T *ptr ) { m_ptr = ptr; }
        ~autoEmpty() { releaseResource(); }

        autoEmpty& operator =( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        autoEmpty& set( T *ptr ) { releaseResource(); m_ptr = ptr; return *this; }
        T         *get() { return m_ptr; }
        T         *detach() { T *p = m_ptr; m_ptr = NULL; return p; }
        void       releaseResource() { if( m_ptr ) { m_ptr->Empty(); m_ptr = NULL; } }

        operator T       *()       { return m_ptr; }
        operator const T *() const { return m_ptr; }
        T&       operator *()       { return *m_ptr; }
        const T& operator *() const { return *m_ptr; }

    protected :
        autoEmpty( const autoEmpty& src );
        autoEmpty& operator =( const autoEmpty& src );
};

template< class T >
class resetVar
{
    public :
        T *m_ptr;
        T  m_value;     // value used to reset the var held in m_ptr

    public :
        resetVar() { m_ptr = NULL; }
        resetVar( T& var, T value ) { m_ptr = &var; m_value = value; }
        ~resetVar() { releaseResource(); }

        resetVar& set( T& var, T value ) { releaseResource(); m_ptr = &var; m_value = value; return *this; }
        T&        get()       { return *m_ptr; }
        const T&  get() const { return *m_ptr; }
        T        *detach() { T *p = m_ptr; m_ptr = NULL; return p; }
        void      releaseResource() { if( m_ptr ) { *m_ptr = m_value; m_ptr = NULL; } }

        operator T       *()       { return m_ptr; }
        operator const T *() const { return m_ptr; }
        T&       operator *()       { return *m_ptr; }
        const T& operator *() const { return *m_ptr; }

    protected :
        resetVar( const resetVar& src );
        resetVar& operator =( const resetVar& src );
};




}

#endif
